'''
Module used to setup the default logging and messaging system.
The system contains on a python logging based approach with logfile,
console output and GUI output dependent on startup options and
message logLevel.
'''

import sys
import atexit
import logging
import logging.handlers
import inspect
from numpy import seterr, seterrcall
from io import StringIO
from threading import Thread
from multiprocessing import Queue, current_process
from queue import Empty
from ..version import __version__ as str_version


# default options used if nothing is set in the configuration
CONSOLE_LEVEL, FILE_LEVEL, GUI_LEVEL = logging.WARNING, logging.DEBUG, logging.INFO

# set log levels according to options
if 'pdb' in list(sys.modules.keys()) or 'pydevd' in list(sys.modules.keys()):
    # if common debugger modules have been loaded, assume a debug run
    CONSOLE_LEVEL, FILE_LEVEL, GUI_LEVEL = logging.INFO, logging.DEBUG, logging.INFO
elif '--debug' in sys.argv:
    CONSOLE_LEVEL, FILE_LEVEL, GUI_LEVEL = logging.DEBUG, logging.DEBUG, logging.INFO


def genx_exit_message():
    if mp_logger:
        mp_logger.join()
    logging.info('*** GenX %s Logging ended ***'%str_version)


# noinspection PyUnusedLocal
def iprint(*objects, sep=None, end=None, file=None, flush=False):
    """
    A logging function that behaves like print but uses logging.info.
    """
    if sep is None:
        sep = ' '
    if end is None:
        end = ''  # '\n'
    logging.info(sep.join(map(str, objects))+end, stacklevel=3)


class NumpyLogger(logging.getLoggerClass()):
    '''
      A logger that makes sure the actual function definition filename, lineno and function name
      is used for logging numpy floating point errors, not the numpy_logger function.
    '''

    def makeRecord(self, name, level, fn, lno, msg, args, exc_info, func=None, extra=None, sinfo=None):
        try:
            curframe = inspect.currentframe()
            calframes = inspect.getouterframes(curframe, 2)
            # stack starts with:
            # (this method, debug call, debug call root logger, numpy_logger, actual function, ...)
            ignore, fname, lineno, func, ignore, ignore = calframes[4]
        except Exception:
            fname = 'unknown'
            lineno = -1
            func = 'unknown'
        return logging.getLoggerClass().makeRecord(self, name, level, fname, lineno,
                                                   msg, args, exc_info, func=func, extra=extra)


nplogger:logging.Logger = None


def numpy_logger(err, _flag):
    nplogger.debug('numpy floating point error encountered (%s)'%err)


def numpy_set_options():
    seterr(divide='call', over='call', under='ignore', invalid='call')
    seterrcall(numpy_logger)

class MPLoggerThread(Thread):
    """
    Performs logging of sub-process started with multiprocessing.
    The initialization routine has to call setup_mp(queue) to use this
    loggers Queue.
    """
    def __init__(self):
        super().__init__(name='MPLogger Receiver')
        self.daemon = True
        self.queue = Queue()
        self.stop_thread = False
        logging.debug(f'Created MPLoggerThread for receiving Queued messages')

    def run(self):
        logging.debug(f'MPLoggerThread started')
        while not self.stop_thread:
            try:
                record = self.queue.get(True, 0.1)
            except Empty:
                pass
            except Exception:
                logging.warning('Error in MPLoggerThread', exc_info=True)
            else:
                if isinstance(record, str):
                    logging.debug(f'MPLogger received string message: {record}')
                else:
                    try:
                        logger = logging.getLogger()
                        logger.handle(record)
                    except Exception:
                        logging.warning('Error in MPLoggerThread record handling', exc_info=True)

    def join(self, timeout=None):
        self.stop_thread = True
        super().join(timeout)

mp_logger:MPLoggerThread = None


def setup_mp(queue: Queue):
    name = current_process().name
    queue.put(f'Start setting up logging in {name}')
    # Called in initialization of new process to allow queued logging
    h = logging.handlers.QueueHandler(queue)  # Just the one handler needed
    root = logging.getLogger()
    root.addHandler(h)
    root.setLevel(logging.DEBUG)

    logging.getLogger('matplotlib').setLevel(logging.WARNING)
    logging.getLogger('numba').setLevel(logging.WARNING)

    logging.debug(f"Activated logging in process {name} using queue handler")
    activate_excepthook()


def setup_system():
    logger = logging.getLogger()
    logger.setLevel(min(CONSOLE_LEVEL, GUI_LEVEL))

    # no console logger for windows (win32gui)
    console = logging.StreamHandler(sys.__stdout__)
    formatter = logging.Formatter('%(levelname) 7s: %(message)s')
    console.setFormatter(formatter)
    console.setLevel(CONSOLE_LEVEL)
    logger.addHandler(console)

    logging.getLogger('matplotlib').setLevel(logging.WARNING)
    logging.getLogger('numba').setLevel(logging.WARNING)
    logging.info(f'*** GenX {str_version} Logging started ***')

    # create MP logger
    global mp_logger
    try:
        mp_logger=MPLoggerThread()
    except Exception:
        logging.warning("Could not start multiprocessing logger:", exc_info=True)
    else:
        mp_logger.start()

    # define numpy warning behavior
    global nplogger
    old_class = logging.getLoggerClass()
    logging.setLoggerClass(NumpyLogger)
    nplogger = logging.getLogger('numpy')
    nplogger.setLevel(logging.DEBUG)
    null_handler = logging.StreamHandler(StringIO())
    null_handler.setLevel(logging.CRITICAL)
    nplogger.addHandler(null_handler)
    logging.setLoggerClass(old_class)
    logging.captureWarnings(True)
    numpy_set_options()

    # write information on program exit
    atexit.register(genx_exit_message)


def activate_logging(logfile):
    logger = logging.getLogger()
    logfile = logging.FileHandler(logfile, 'w', encoding='utf-8')
    logger.setLevel(min(logger.getEffectiveLevel(), FILE_LEVEL))
    formatter = logging.Formatter('[%(levelname)s] - %(asctime)s - '
                                  '%(process)d:%(threadName)s:%(filename)s:%(lineno)i:%(funcName)s '
                                  '| %(message)s',
                                  '')
    logfile.setFormatter(formatter)
    logfile.setLevel(FILE_LEVEL)
    logger.addHandler(logfile)
    nplogger.addHandler(logfile)
    logger.info('*** GenX %s Logging started to file ***'%str_version)


_prev_excepthook = None


def excepthook_overwrite(*exc_info):
    # making sure all exceptions are displayed on GUI and logged
    # noinspection PyTypeChecker
    logging.critical('uncought python error', exc_info=exc_info)
    # noinspection PyCallingNonCallable
    return _prev_excepthook(*exc_info)


def activate_excepthook():
    logging.debug("replacing sys.excepthook with user function")
    global _prev_excepthook
    _prev_excepthook = sys.excepthook
    sys.excepthook = excepthook_overwrite



"""
Native WX replacement for the matplotlib plot panel.
"""
import wx
import numpy as np
import threading
from dataclasses import dataclass
from typing import Type
from logging import debug, getLogger, ERROR
from wx.lib import plot as wxplot

from .custom_events import plot_position, state_changed, skips_event
from ..core.config import BaseConfig, Configurable
from ..model import Model
from ..data import DataList


MARKER_STYLES = {
    'o': 'circle',
    's': 'square',
    '.': 'dot',
    'd': 'triangle_down',
    '<': 'triangle',
    None: 'dot',
    }

LINE_STYLES = {
    '-': wx.PENSTYLE_SOLID,
    ':': wx.PENSTYLE_DOT,
    '--': wx.PENSTYLE_LONG_DASH,
    '.-': wx.PENSTYLE_DOT_DASH,
    None: wx.PENSTYLE_INVALID,
    }

@dataclass
class BasePlotConfig(BaseConfig):
    zoom: bool = False
    autoscale: bool = True
    x_scale: str = 'linear'
    y_scale: str = 'linear'

class PlotPanel(wx.Panel, Configurable):
    '''
    Base class for the plotting in GenX - all the basic functionallity
    should be implemented in this class. The plots should be derived from
    this class. These classes should implement an update method to update
    the plots.
    '''
    opt: BasePlotConfig

    def __init__(self, parent, id=-1, color=None, dpi=None, config_class: Type[BasePlotConfig] = None,
                 style=wx.NO_FULL_REPAINT_ON_RESIZE | wx.EXPAND | wx.ALL, **kwargs):
        debug('start init PlotPanel')
        wx.Panel.__init__(self, parent, id=id, style=style, **kwargs)
        Configurable.__init__(self, config_class)
        if dpi is None:
            dpi = wx.GetApp().dpi_scale_factor*96.  # wx.GetDisplayPPI()[0]
        self.parent = parent
        self._last_graph = None
        self._initial_scale = True

        # Flags and bindings for zooming
        self.opt.load_config()
        self._x_range = None
        self._y_range = None
        self.zooming = False
        self.canvas = wxplot.PlotCanvas(self)
        self.canvas.enableAntiAliasing = True

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(self.canvas, 1, wx.EXPAND)
        self.SetSizer(sizer)
        self.OnPlotDraw(None)
        self.canvas.pointLabelFunc = self.DrawPointLabel

        self.canvas.Bind(wx.EVT_MOUSEWHEEL, self.OnMouseScroll)
        self.canvas.canvas.Bind(wx.EVT_MOTION, self.OnMouseMove)
        self.canvas.canvas.Bind(wx.EVT_RIGHT_UP, self.OnContextMenu)
        self.canvas.enablePointLabel=True

    def OnPlotDraw(self, event):
        """ Sin, Cos, and Points """
        self.resetDefaults()
        self.update()

    def SetCallbackWindow(self, window):
        self.callback_window = window

    def SetAutoScale(self, do_autoscale):
        self.opt.autoscale = do_autoscale
        if self._last_graph:
            self.do_replot()

    def GetAutoScale(self):
        return self.opt.autoscale

    def SetZoom(self, active=False):
        self.opt.zoom = active
        if active:
            self.canvas.enableZoom = True
            self.canvas.enablePointLabel = False
        else:
            self.canvas.enableZoom = False
            self.canvas.enablePointLabel = True

    def GetZoom(self):
        return self.opt.zoom

    def GetYScale(self):
        return self.opt.y_scale

    def GetXScale(self):
        return self.opt.x_scale

    def UpdateConfigValues(self):
        self.SetZoom(self.opt.zoom)
        self.SetAutoScale(self.opt.autoscale)

    def ReadConfig(self):
        self._initial_scale = True
        super().ReadConfig()

    def ExecuteAutoscale(self):
        if not self.opt.autoscale and not self._initial_scale and self.opt.zoom:
            try:
                self._x_range = tuple(self.canvas.xCurrentRange)
            except Exception as e:
                pass
            try:
                self._y_range = tuple(self.canvas.yCurrentRange)
            except Exception:
                pass
        elif self.opt.autoscale or self._initial_scale:
            self._x_range = None
            self._y_range = None
            self._initial_scale = False
        self.canvas.logScale = (self.opt.x_scale == 'log', self.opt.y_scale == 'log')
        if self._x_range:
            if self.canvas.logScale[0]:
                self.canvas.xSpec=map(np.log10, self._x_range)
            else:
                self.canvas.xSpec=self._x_range
        else:
            self.canvas.xSpec = 'auto'
        if self._y_range:
            if self.canvas.logScale[1]:
                self.canvas.ySpec=tuple(map(np.log10, self._y_range))
            else:
                self.canvas.ySpec=self._y_range
        else:
            self.canvas.ySpec = 'auto'

    def resetDefaults(self):
        """Just to reset the fonts back to the PlotCanvas defaults"""
        self.canvas.SetFont(wx.Font(10,
                                    wx.FONTFAMILY_SWISS,
                                    wx.FONTSTYLE_NORMAL,
                                    wx.FONTWEIGHT_NORMAL)
                            )
        self.canvas.fontSizeAxis = 10
        self.canvas.fontSizeLegend = 7
        self.ExecuteAutoscale()

    def plot_result(self, graph, delayed=False):
        self._to_plot = graph
        if threading.main_thread() and not delayed:
            self.do_plot_result()
        else:
            wx.CallLater(10, self.do_plot_result)

    def do_plot_result(self):
        if self._to_plot is None:
            return
        self._last_graph = self._to_plot
        self._to_plot = None
        self.do_replot()

    def do_replot(self):
        self.ExecuteAutoscale()
        self.canvas.Draw(self._last_graph)
        self.Refresh()

    def update(self, data=None):
        pass

    def get_colour(self, float_colors):
        int_colors=[int(ci*255) for ci in float_colors]
        return wx.Colour(*int_colors)

    def OnMouseScroll(self, event: wx.MouseEvent):
        if event.GetWheelAxis()==wx.MOUSE_WHEEL_HORIZONTAL:
            event.Skip()
            return
        rot = event.GetWheelRotation()/120.
        if event.ControlDown():
            rot *= 0.1
        if event.AltDown():
            # horizontal scaling
            if self._x_range:
                xmin, xmax = self._x_range
            else:
                xmin, xmax = self.canvas.xCurrentRange
            xrange = xmax-xmin
            if event.ShiftDown():
                if self.opt.x_scale=='log':
                    if rot>0:
                        self._x_range=(xmin*(1+2.33333*rot), xmax)
                    else:
                        self._x_range=(xmin/(1-2.33333*rot), xmax)
                else:
                    self._x_range=(xmin+xrange*0.2*rot, xmax)
            else:
                if self.opt.x_scale=='log':
                    if rot>0:
                        self._x_range=(xmin, xmax*(1+2.33333*rot))
                    else:
                        self._x_range=(xmin, xmax/(1-2.33333*rot))
                else:
                    self._x_range=(xmin, xmax+xrange*0.2*rot)
        else:
            # vertical scaling
            if self._y_range:
                ymin, ymax = self._y_range
            else:
                ymin, ymax = self.canvas.yCurrentRange
            yrange = ymax-ymin
            if event.ShiftDown():
                if self.opt.y_scale=='log':
                    if rot>0:
                        self._y_range=(ymin*(1+2.33333*rot), ymax)
                    else:
                        self._y_range=(ymin/(1-2.33333*rot), ymax)
                else:
                    self._y_range=(ymin+yrange*0.2*rot, ymax)
            else:
                if self.opt.y_scale=='log':
                    if rot>0:
                        self._y_range=(ymin, ymax*(1+2.33333*rot))
                    else:
                        self._y_range=(ymin, ymax/(1-2.33333*rot))
                else:
                    self._y_range=(ymin, ymax+yrange*0.2*rot)
        self.opt.autoscale = False
        tmp = self.opt.zoom
        self.opt.zoom = False
        self.do_replot()
        self.opt.zoom = tmp

    def OnMouseMove(self, event: wx.MouseEvent):
        if self.canvas.enablePointLabel:
            px, py = self.canvas.GetXY(event)
            if self.canvas.logScale[0]:
                px = np.log10(px)
            if self.canvas.logScale[1]:
                py = np.log10(py)
            dlst = self.canvas.GetClosestPoint((px, py), pointScaled=True)
            if dlst != []:
                curveNum, legend, pIndex, pointXY, scaledXY, distance = dlst
                px, py = pointXY
                if self.canvas.logScale[0]:
                    px = 10**px
                if self.canvas.logScale[1]:
                    py = 10**py
                mDataDict = {"curveNum": curveNum,
                             "legend": legend,
                             "pIndex": pIndex,
                             "pointXY": (px, py),
                             "scaledXY": scaledXY}
                self.canvas.UpdatePointLabel(mDataDict)
        event.Skip()

    def DrawPointLabel(self, dc, mDataDict):
        xmin, xmax = self.canvas.xCurrentRange
        ymin, ymax = self.canvas.yCurrentRange

        if self.canvas.logScale[0]:
            xmax = np.log10(xmax)
        if self.canvas.logScale[1]:
            ymax = np.log10(ymax)
        right, top = self.canvas.PositionUserToScreen((xmax, ymax))

        dc.SetPen(wx.Pen(wx.BLACK))
        dc.SetBrush(wx.Brush(wx.BLACK, wx.BRUSHSTYLE_SOLID))

        sx, sy = mDataDict["scaledXY"]  # scaled x,y of closest point
        # 10by10 square centered on point
        dc.DrawRectangle(sx - 5, sy - 5, 10, 10)
        px, py = mDataDict["pointXY"]
        cNum = mDataDict["curveNum"]
        pntIn = mDataDict["pIndex"]
        legend = mDataDict["legend"]
        # make a string to display
        txt=f"x={px:.6g}\ny={py:.6g}"
        txt_width = max([dc.GetTextExtent(ti)[0] for ti in txt.splitlines()])
        dc.DrawText(txt, int(right)-10-txt_width, int(top)+10)

    def OnContextMenu(self, event):
        menu = self.generate_context_menu()
        self.PopupMenu(menu)
        self.Unbind(wx.EVT_MENU)
        menu.Destroy()

    def generate_context_menu(self):
        menu = wx.Menu()
        zoomID = wx.NewId()
        menu.AppendCheckItem(zoomID, "Zoom")
        menu.Check(zoomID, self.GetZoom())

        def OnZoom(event):
            self.SetZoom(not self.GetZoom())

        self.Bind(wx.EVT_MENU, OnZoom, id=zoomID)
        zoomallID = wx.NewId()
        menu.Append(zoomallID, 'Zoom All')

        def zoomall(event):
            tmp = self.opt.autoscale
            self.SetAutoScale(True)
            self.opt.autoscale = tmp

        self.Bind(wx.EVT_MENU, zoomall, id=zoomallID)
        copyID = wx.NewId()
        menu.Append(copyID, "Copy")

        def copy(event):
            context = wx.ClientDC(self.canvas.canvas)
            memory = wx.MemoryDC()
            x, y = self.canvas.canvas.ClientSize
            bitmap = wx.Bitmap(x, y, -1)
            memory.SelectObject(bitmap)
            memory.Blit(0, 0, x, y, context, 0, 0)
            memory.SelectObject(wx.NullBitmap)
            bmp_obj = wx.BitmapDataObject()
            bmp_obj.SetBitmap(bitmap)

            if not wx.TheClipboard.IsOpened():
                open_success = wx.TheClipboard.Open()
                if open_success:
                    wx.TheClipboard.SetData(bmp_obj)
                    wx.TheClipboard.Close()
                    wx.TheClipboard.Flush()


        menu.AppendSeparator()

        self.Bind(wx.EVT_MENU, copy, id=copyID)
        yscalemenu = wx.Menu()
        logID = wx.NewId()
        linID = wx.NewId()
        yscalemenu.AppendRadioItem(logID, "log")
        yscalemenu.AppendRadioItem(linID, "linear")
        menu.Append(-1, "y-scale", yscalemenu)
        if self.opt.y_scale=='log':
            yscalemenu.Check(logID, True)
        else:
            yscalemenu.Check(linID, True)

        def yscale_log(event):
            self.opt.y_scale = 'log'
            self.do_replot()

        def yscale_lin(event):
            self.opt.y_scale = 'linear'
            self.do_replot()

        self.Bind(wx.EVT_MENU, yscale_log, id=logID)
        self.Bind(wx.EVT_MENU, yscale_lin, id=linID)
        xscalemenu = wx.Menu()
        logID = wx.NewId()
        linID = wx.NewId()
        xscalemenu.AppendRadioItem(logID, "log")
        xscalemenu.AppendRadioItem(linID, "linear")
        menu.Append(-1, "x-scale", xscalemenu)
        if self.opt.x_scale=='log':
            xscalemenu.Check(logID, True)
        else:
            xscalemenu.Check(linID, True)

        def xscale_log(event):
            self.opt.x_scale = 'log'
            self.do_replot()

        def xscale_lin(event):
            self.opt.x_scale = 'linear'
            self.do_replot()

        self.Bind(wx.EVT_MENU, xscale_log, id=logID)
        self.Bind(wx.EVT_MENU, xscale_lin, id=linID)
        autoscaleID = wx.NewId()
        menu.AppendCheckItem(autoscaleID, "Autoscale")
        menu.Check(autoscaleID, self.GetAutoScale())

        def OnAutoScale(event):
            self.SetAutoScale(not self.GetAutoScale())

        self.Bind(wx.EVT_MENU, OnAutoScale, id=autoscaleID)
        return menu

class DataPanelConfig(BasePlotConfig):
    section = 'data plot'


class DataPlotPanel(PlotPanel):
    ''' Class for plotting the data and the fit
    '''
    _last_poptions = None

    def __init__(self, parent, id=-1, color=None, dpi=None
                 , style=wx.NO_FULL_REPAINT_ON_RESIZE, **kwargs):
        self.main_ax_rect = (0.125, 0.3, 0.8, 0.6)
        self.sub_ax_rect = (0.125, 0.1, 0.8, 0.18)
        PlotPanel.__init__(self, parent, id, color, dpi, DataPanelConfig, style, **kwargs)
        sizer = self.GetSizer()
        sizer.GetItem(0).SetProportion(3)

        self.canvas_fom = wxplot.PlotCanvas(self)
        self.canvas_fom.enableAntiAliasing = True
        sizer.Add(self.canvas_fom, 1, wx.EXPAND)

        self.update = self.plot_data
        self._last_xlabel = 'x'
        self._last_ylabel = 'y'
        self.resetDefaultsFOM()

    def resetDefaultsFOM(self):
        self.canvas_fom.SetFont(wx.Font(10,
                                    wx.FONTFAMILY_SWISS,
                                    wx.FONTSTYLE_NORMAL,
                                    wx.FONTWEIGHT_NORMAL)
                            )
        self.canvas_fom.fontSizeAxis = 10
        self.canvas_fom.fontSizeLegend = 7
        self.canvas_fom.enableAxesValues = (False, True)

    def ExecuteAutoscale(self):
        if getattr(self, 'canvas_fom', None):
            self.canvas_fom.logScale = (self.canvas.logScale[0], False)
            self.canvas_fom.xSpec = tuple(self.canvas.xCurrentRange)
            self.canvas_fom.ySpec = 'auto'
            if not self.opt.autoscale and self._x_range and self.canvas.logScale[0]:
                        self.canvas.xSpec=map(np.log10, self.canvas.xCurrentRange)

    def plot_result(self, graph, fom_graph, delayed=False):
        self._to_fom = fom_graph
        super().plot_result(graph, delayed=delayed)

    def do_plot_result(self):
        if self._to_plot is None:
            return
        self._last_graph = (self._to_plot, self._to_fom)
        self._to_plot = None
        self._to_fom = None
        self.do_replot()

    def do_replot(self):
        super().ExecuteAutoscale()
        self.canvas.Draw(self._last_graph[0])
        if self._last_graph[1]:
            self.ExecuteAutoscale()
            self.canvas_fom.Draw(self._last_graph[1])
        self.Refresh()

    def get_data_plots(self, data:DataList):
        lines = []
        p_datasets = [data_set for data_set in data if data_set.show]
        pe_datasets = [data_set for data_set in data if data_set.use_error and data_set.show]

        for data_set in p_datasets:
            data1 = np.vstack([data_set.x, np.nan_to_num(data_set.y)]).T
            if data_set.data_linetype!='':
                item = wxplot.PolyLine(data1,
                                       colour=self.get_colour(data_set.data_color),
                                       width=data_set.data_linethickness,
                                       style=LINE_STYLES[data_set.data_linetype],
                                       drawstyle='line',
                                       )
                lines.append(item)
            if data_set.data_symbol!='':
                item = wxplot.PolyMarker(data1,
                                         colour=self.get_colour(data_set.data_color),
                                         fillcolour=self.get_colour(data_set.data_color),
                                         marker=MARKER_STYLES[data_set.data_symbol],
                                         width=data_set.data_linethickness,
                                         size=data_set.data_symbolsize*0.2,
                                         fillstyle=wx.BRUSHSTYLE_SOLID,
                                         )
                lines.append(item)

            if data_set in pe_datasets:
                ylow = np.nan_to_num(data_set.y)-np.nan_to_num(data_set.error)
                yhigh = np.nan_to_num(data_set.y)+np.nan_to_num(data_set.error)
                for xi, li, hi in zip(data_set.x, ylow, yhigh):
                    data1 = [[xi, li], [xi, hi]]
                    item = wxplot.PolyLine(data1,
                                           colour=self.get_colour(data_set.data_color),
                                           width=data_set.data_linethickness,
                                           style=wx.PENSTYLE_SOLID,
                                           drawstyle='line',
                                           )
                    lines.append(item)
        return lines

    def get_sim_lines(self, data:DataList):
        lines = []
        p_datasets = [data_set for data_set in data if data_set.show]

        for data_set in p_datasets:
            data1 = np.vstack([data_set.x, np.nan_to_num(data_set.y_sim)]).T
            if data_set.sim_linetype!='':
                item = wxplot.PolyLine(data1,
                            colour=self.get_colour(data_set.sim_color),
                            width=data_set.sim_linethickness,
                            style=LINE_STYLES[data_set.sim_linetype],
                            drawstyle='line',
                            )
                lines.append(item)
            if data_set.sim_symbol!='':
                item = wxplot.PolyMarker(data1,
                                       colour=self.get_colour(data_set.sim_color),
                                       fillcolour=self.get_colour(data_set.sim_color),
                                       marker=MARKER_STYLES[data_set.sim_symbol],
                                       width=data_set.sim_linethickness,
                                       size=data_set.sim_symbolsize*0.2,
                                       fillstyle=wx.BRUSHSTYLE_SOLID,
                                       )

                lines.append(item)
        return lines

    def get_fom_lines(self, data:DataList):
        lines = []
        p_datasets = [data_set for data_set in data if data_set.show]

        for data_set in p_datasets:
            data1 = np.vstack([data_set.x, np.nan_to_num(data_set.y_fom)]).T
            if data_set.sim_linetype!='':
                item = wxplot.PolyLine(data1,
                            colour=self.get_colour(data_set.sim_color),
                            width=data_set.sim_linethickness,
                            style=LINE_STYLES[data_set.sim_linetype],
                            drawstyle='line',
                            )
                lines.append(item)
            if data_set.sim_symbol!='':
                item = wxplot.PolyMarker(data1,
                                       colour=self.get_colour(data_set.sim_color),
                                       fillcolour=self.get_colour(data_set.sim_color),
                                       marker=MARKER_STYLES[data_set.sim_symbol],
                                       width=data_set.sim_linethickness,
                                       size=data_set.sim_symbolsize*0.2,
                                       fillstyle=wx.BRUSHSTYLE_SOLID,
                                       )

                lines.append(item)
        return lines

    def plot_data(self, data: DataList, xlabel=None, ylabel=None):
        if xlabel is not None:
            self._last_xlabel = xlabel.replace('$^{-1}$', '⁻¹')
        if ylabel is not None:
            self._last_ylabel = ylabel.replace('$^{-1}$', '⁻¹')
        ax1_lines = self.get_data_plots(data)
        self.plot_result(wxplot.PlotGraphics(ax1_lines, "", xLabel=self._last_xlabel, yLabel=self._last_ylabel),
                         None)

    def plot_data_fit(self, data: DataList, xlabel=None, ylabel=None):
        return self.plot_data_sim(data, xlabel, ylabel)

    def plot_data_sim(self, data:DataList, xlabel=None, ylabel=None, delayed=False):
        if xlabel is not None:
            self._last_xlabel = xlabel.replace('$^{-1}$', '⁻¹')
        if ylabel is not None:
            self._last_ylabel = ylabel.replace('$^{-1}$', '⁻¹')
        ax1_lines = self.get_data_plots(data)+self.get_sim_lines(data)
        ax2_lines = self.get_fom_lines(data)
        self.plot_result(wxplot.PlotGraphics(ax1_lines, "", xLabel=self._last_xlabel, yLabel=self._last_ylabel),
                         wxplot.PlotGraphics(ax2_lines, "", xLabel="", yLabel="FOM"),
                         delayed=delayed)

    @skips_event
    def OnDataListEvent(self, event):
        data_list = event.GetData()
        if event.data_changed:
            self.plot_data(data_list)

    @skips_event
    def OnSimPlotEvent(self, event):
        model: Model = event.GetModel()
        data_list = model.get_data()
        try:
            ylabel = model.eval_in_model('globals().get("__ylabel__", getattr(model, "__ylabel__", "y"))')
        except NameError:
            ylabel = model.eval_in_model('globals().get("__ylabel__", "y")')
        try:
            xlabel = model.eval_in_model('globals().get("__xlabel__", getattr(model, "__xlabel__", "x"))')
        except NameError:
            xlabel = model.eval_in_model('globals().get("__xlabel__", "x")')
        self.plot_data_sim(data_list, xlabel=xlabel, ylabel=ylabel)

    @skips_event
    def OnSolverPlotEvent(self, event):
        if event.update_fit:
            self.plot_data_sim(event.data, delayed=True)


class ErrorPanelConfig(BasePlotConfig):
    section = 'fom plot'


class ErrorPlotPanel(PlotPanel):
    '''
    Class for plotting evolution of the error as a function of the generations.
    '''

    def __init__(self, parent, id=-1, color=None, dpi=None
                 , style=wx.NO_FULL_REPAINT_ON_RESIZE, **kwargs):
        PlotPanel.__init__(self, parent, id, color, dpi, ErrorPanelConfig, style, **kwargs)
        self.update = self.errorplot
        self.update(None)

    def errorplot(self, data):
        if data is None:
            theta = np.arange(0.1, 10, 0.001)
            data1 = np.vstack([theta, np.floor(15-theta)]).T
        else:
            data1 = np.asarray(data)

        item = wxplot.PolyLine(data1,
                               colour='red',
                               width=2,
                               style=wx.PENSTYLE_SOLID,
                               drawstyle='line',
                               )
        self.plot_result(wxplot.PlotGraphics([item], "", xLabel='Iteration', yLabel='FOM'), delayed=True)

    @skips_event
    def OnSolverPlotEvent(self, event):
        '''
        Event handler function to connect to solver update events i.e.
        update the plot with the simulation
        '''
        fom_log = event.fom_log
        self.update(fom_log)


class ParsPanelConfig(BasePlotConfig):
    section = 'pars plot'


class ParsPlotPanel(PlotPanel):
    '''
    Class to plot the diffrent parametervalus during a fit.
    '''

    def __init__(self, parent, id=-1, color=None, dpi=None
                 , style=wx.NO_FULL_REPAINT_ON_RESIZE, **kwargs):
        self.auto_x_range = (-0.5, 1.5)
        self.auto_y_range = (0., 1.)
        PlotPanel.__init__(self, parent, id, color, dpi, ParsPanelConfig, style, **kwargs)
        self.update = self.Plot
        self.update(None)

    def Plot(self, data):
        '''
        Plots each variable and its max and min value in the population.
        '''
        if data is None or not data.fitting:
            pars = np.arange(10)
            pdata = np.vstack([pars, pars*0.01+0.4]).T
            pop_min = pdata[:, 1]-0.05
            pop_max = pdata[:, 1]+0.05
        else:
            pop = np.array(data.population)
            norm = 1.0/(data.max_val-data.min_val)
            best = (np.array(data.values)-data.min_val)*norm
            pop_min = (pop.min(0)-data.min_val)*norm
            pop_max = (pop.max(0)-data.min_val)*norm

            pars = np.arange(len(best))
            pdata = np.vstack([pars, best]).T

        points = wxplot.PolyMarker(pdata,
                                 colour='red',
                                 fillcolour='red',
                                 marker='circle',
                                 width=2.0,
                                 size=1.5,
                                 fillstyle=wx.BRUSHSTYLE_SOLID,
                                 )
        bars = []
        for xi, mini, maxi in zip(pars, pop_min, pop_max):
            points1 = [(xi, mini), (xi, maxi)]
            bars.append(wxplot.PolyLine(points1, colour='blue', width=25))

        self.auto_x_range = (-0.5, pars[-1]+0.5)
        self.auto_y_range = (0., 1.)
        self.plot_result(wxplot.PlotGraphics(bars+[points], "",
                                             xLabel='Parameter Index (only fittable)',
                                             yLabel='Relative value in min/max range'),
                         delayed=True)

    def ExecuteAutoscale(self):
        # Allows to overwrite the behavior when autoscale is active
        self.canvas.xSpec = self.auto_x_range
        self.canvas.ySpec = self.auto_y_range
        if not self.opt.autoscale:
            if self._x_range:
                self.canvas.xSpec=self._x_range
            if self._y_range:
                self.canvas.ySpec=self._y_range

    @skips_event
    def OnSolverParameterEvent(self, event):
        '''
        Event handler function to connect to solver update events i.e.
        update the plot during the fitting
        '''
        self.update(event)


class FomPanelConfig(BasePlotConfig):
    section = 'fom scan plot'


class FomScanPlotPanel(PlotPanel):
    '''
    Class to take care of fom scans.
    '''

    def __init__(self, parent, id=-1, color=None, dpi=None
                 , style=wx.NO_FULL_REPAINT_ON_RESIZE, **kwargs):
        self.type = 'project'
        PlotPanel.__init__(self, parent, id, color, dpi, FomPanelConfig, style, **kwargs)
        self.update = self.Plot
        self.update(None)

    def SetPlottype(self, type):
        '''SetScantype(self, type) --> None

        Sets the type of the scan type = "project" or "scan"
        '''
        if type.lower()=='project':
            self.type = 'project'
            self.opt.autoscale = False
        elif type.lower()=='scan':
            self.type = 'scan'
            self.opt.autoscale = True

    def Plot(self, data, l1='Parameter', l2='FOM'):
        '''
        Plots each variable and its max and min value in the population.
        '''
        if data is None:
            return self.plot_result(wxplot.PlotGraphics([wxplot.PolyLine([(-1, 0.1), (1, 0.1)])], "", xLabel=l1, yLabel=l2))
        x, y, bestx, besty, e_scale = data[:5]
        if self.type.lower()=='project':
            pdata = np.vstack([x,y]).T
            points = wxplot.PolyMarker(pdata,
                                       colour='blue',
                                       fillcolour='blue',
                                       marker='circle',
                                       width=2.0,
                                       size=1.5,
                                       fillstyle=wx.BRUSHSTYLE_SOLID,
                                       )
            bpoint = wxplot.PolyMarker([(bestx, besty)],
                                       colour='red',
                                       fillcolour='red',
                                       marker='circle',
                                       width=2.0,
                                       size=1.5,
                                       fillstyle=wx.BRUSHSTYLE_SOLID,
                                       )
            line = wxplot.PolyLine([(x.min(), besty*e_scale), (x.max(), besty*e_scale)], colour='red', width=1)
            items = [points, line, bpoint]

            self._x_range = (x.min(), x.max())
            self._y_range = (min(y.min(), besty)*0.95, (besty*e_scale-min(y.min(), besty))*2.0+min(y.min(), besty))
        elif self.type.lower()=='scan':
            pdata = np.vstack([x,y]).T
            points = wxplot.PolyLine(pdata, colour='blue', width=1)
            bpoint = wxplot.PolyMarker([(bestx, besty)],
                                       colour='red',
                                       fillcolour='red',
                                       marker='circle',
                                       width=2.0,
                                       size=1.5,
                                       fillstyle=wx.BRUSHSTYLE_SOLID,
                                       )
            line = wxplot.PolyLine([(x.min(), besty*e_scale), (x.max(), besty*e_scale)], colour='red', width=1)
            items = [points, line, bpoint]

            self._x_range = (x.min(), x.max())
            self._y_range = (min(y.min(), besty)*0.95, y.max()*1.05)

        self.plot_result(wxplot.PlotGraphics(items, "", xLabel=l1, yLabel=l2))



import os
import sys
import appdirs
import numba.core.caching as nc

from logging import debug
from ...version import __version__ as program_version

config_path = os.path.abspath(appdirs.user_data_dir('GenX3', 'ArturGlavic'))

class _GenxCacheLocator(nc._SourceFileBackedLocatorMixin, nc._CacheLocator):
    """
    A locator that points to the GenX specific numba cache directory.
    It dempents on the genx version but not the location of the source files.
    """

    def __init__(self, py_func, py_file):
        self._py_file = py_file
        self._lineno = py_func.__code__.co_firstlineno
        cache_subpath = self.get_suitable_cache_subpath(py_file)
        base_dir = os.environ.get('NUMBA_CACHE_DIR', os.path.join(config_path, 'numba_cache'))
        self._cache_path = os.path.join(base_dir, cache_subpath)
        debug(f'numba cache path for {py_func} from {py_file} is {self._cache_path}')
        debug(f'    source stamp is {self.get_source_stamp()}')

    def get_source_stamp(self):
        # overwrite numba behavior to make sure the actual file times are used and not the executable
        st = os.stat(self._py_file)
        return st.st_mtime, st.st_size

    def get_cache_path(self):
        return self._cache_path

    @classmethod
    def from_function(cls, py_func, py_file):
        parent = super(_GenxCacheLocator, cls)
        if py_file.startswith(os.path.join('genx', 'models', 'lib')):
            from genx.models import lib
            base_dir = os.path.dirname(os.path.abspath(lib.__file__))
            py_file = os.path.join(base_dir, os.path.basename(py_file))
        output = parent.from_function(py_func, py_file)
        if output is None:
            debug(f'no locator created for file {py_file}, exists={os.path.exists(py_file)}')
            self = _GenxCacheLocator(py_func, py_file)
            try:
                self.ensure_cache_path()
            except OSError:
                debug('OSError when checkint cache path', exc_info=True)
            # Perhaps a placeholder (e.g. "<ipython-XXX>")
            return
        return output

    @classmethod
    def get_suitable_cache_subpath(cls, py_file):
        path = os.path.abspath(py_file)
        subpath = os.path.dirname(path)
        parentdir = os.path.split(subpath)[-1]
        # separate caches for source and binary distribution
        # mostly if source is used to test on the same machine
        if getattr(sys, 'frozen', False):
            prefix = 'genxsource'
        else:
            prefix = 'genxfrozen'
        return f'{prefix}_{program_version.replace(".", "_")}_{parentdir}'


def configure_numba():
    if hasattr(nc, '_CacheImpl'):
        nc._CacheImpl._locator_classes.insert(0, _GenxCacheLocator)
    else:
        # Newer version of numba
        nc.CacheImpl._locator_classes.insert(0, _GenxCacheLocator)



"""
    Test of the instrument module the implements resolution convolution and footprint corrections.
"""
import unittest
import numpy as np
from importlib import reload
from unittest.mock import patch
from scipy.special import erf

from genx.models import lib
lib.USE_NUMBA=False
from genx.models.lib import instrument


class TestFootprint(unittest.TestCase):
    def test_fp_square(self):
        ai=np.linspace(0., 10., 100)
        slen=10.0
        bw=np.sin(5.0*np.pi/180.)*slen # beam with to have full illumination at 5 deg

        IFP=instrument.SquareIntensity(ai, slen, bw)
        self.assertTrue((IFP>=0.0).all())
        np.testing.assert_array_equal(IFP[ai>=5.0], 1.0)
        np.testing.assert_array_less(IFP[ai<5.0], 1.0)

    def test_fp_gauss(self):
        ai=np.linspace(0., 10., 100)
        slen=10.0
        bw=np.sin(5.0*np.pi/180.)*slen/2.0 # beam with to have +/-1 sigma illumination at 5 deg
        one_sigma=erf(1.0/np.sqrt(2.0)) # area of Gaussian between + and - sigma

        IFP=instrument.GaussIntensity(ai, slen/2., slen/2., bw)
        self.assertTrue((IFP>=0.0).all())
        self.assertTrue((IFP<1.0).all())
        np.testing.assert_array_less(one_sigma, IFP[ai>=5.0])
        np.testing.assert_array_less(IFP[ai<5.0], one_sigma)

    def test_fp_gauss(self):
        ai=np.linspace(0., 10., 100)
        slen=10.0
        bw=np.sin(5.0*np.pi/180.)*slen/2.0  # beam with to have +/-1 sigma illumination at 5 deg
        one_sigma=erf(1.0/np.sqrt(2.0))  # area of Gaussian between + and - sigma

        IFP=instrument.GaussIntensity(ai, slen/2., slen/2., bw)
        self.assertTrue((IFP>=0.0).all())
        self.assertTrue((IFP<1.0).all())
        np.testing.assert_array_less(one_sigma, IFP[ai>=5.0])
        np.testing.assert_array_less(IFP[ai<5.0], one_sigma)

    def test_fp_gauss_asym(self):
        ai=np.linspace(0., 10., 100)
        slen=10.0
        bw=np.sin(5.0*np.pi/180.)*slen/2.0  # beam with to have +/-1 sigma illumination at 5 deg
        two_sigma=erf(2.0/np.sqrt(2.0))  # area of Gaussian between + and - 2*sigma
        one_sigma=erf(1.0/np.sqrt(2.0))  # area of Gaussian between + and - sigma
        footprint=(two_sigma+one_sigma)/2.

        IFP=instrument.GaussIntensity(ai, slen, slen/2., bw)
        self.assertTrue((IFP>=0.0).all())
        self.assertTrue((IFP<1.0).all())
        np.testing.assert_array_less(footprint, IFP[ai>=5.0])
        np.testing.assert_array_less(IFP[ai<5.0], footprint)

class TestResolution(unittest.TestCase):
    def test_res_vector(self):
        q_in=np.linspace(0.01, 0.301, 300)
        n_points=5
        dq=0.001
        rng=3.0

        q_out, weights=instrument.ResolutionVector(q_in, dq, n_points, range=rng)
        self.assertEqual(q_in.shape[0]*n_points, q_out.shape[0])
        self.assertEqual((n_points, q_in.shape[0]), weights.shape)
        for i, dqi in enumerate(np.linspace(-rng*dq, rng*dq, n_points)):
            np.testing.assert_array_almost_equal(q_in, q_out.reshape(n_points, -1)[i]-dqi) # center of resolution is q-point

    def test_res_vector_vec(self):
        q_in=np.linspace(0.01, 0.301, 300)
        n_points=5
        dq=0.001
        rng=3.0

        q_out_s, weights_s=instrument.ResolutionVector(q_in, dq, n_points, range=rng)
        q_out_v, weights_v=instrument.ResolutionVector(q_in, dq*np.ones_like(q_in), n_points, range=rng)
        np.testing.assert_array_equal(q_out_s, q_out_v)
        np.testing.assert_array_equal(weights_s, weights_v)

    def test_convolution(self):
        q_in=np.linspace(0.01, 0.301, 300)
        n_points=5

        q_res=np.linspace(-0.0001, 0.0001, n_points)[:, np.newaxis]+q_in
        w_res=1./(n_points-1)*np.ones_like(q_res)
        q_res=q_res.flatten()
        I0=np.ones_like(q_res)

        # 1 intensity everywhere should convolute to 1
        Iconv=instrument.ConvoluteResolutionVector(q_res, I0, w_res)
        np.testing.assert_array_equal(Iconv, 1.0)
        # single 1 value in convolution range should give weight fraction as value
        I0=I0.reshape(n_points, -1)
        I0[np.linspace(-1,1,n_points)!=0]=0
        I0=I0.flatten()
        Iconv=instrument.ConvoluteResolutionVector(q_res, I0, w_res)
        np.testing.assert_array_almost_equal(Iconv, 1.0/(n_points-1))
        # full weight on point with 1 intensity should yield 1 again
        w_res[np.linspace(-1,1,n_points)!=0]=0
        Iconv=instrument.ConvoluteResolutionVector(q_res, I0, w_res)
        np.testing.assert_array_equal(Iconv, 1.0)

    def test_fft_convolve(self):
        q_in=np.linspace(0.01, 0.301, 300)
        I0=np.ones_like(q_in)

        Iconv=instrument.ConvoluteFast(q_in, I0, 0.001, range=3)
        np.testing.assert_array_almost_equal(Iconv, 1.0)

class TestInstrumentModuleBranching(unittest.TestCase):
    def setUp(self):
        from genx.models.lib import instrument_numba
        self.SquareIntensity=instrument_numba.SquareIntensity

    def tearDown(self):
        global instrument
        lib.USE_NUMBA=False
        reload(instrument)
        # make sure to reset the module, in case test_numba_error breaks before reset
        from genx.models.lib import instrument_numba
        instrument_numba.SquareIntensity=self.SquareIntensity

    def test_numba(self):
        global instrument
        lib.USE_NUMBA=True
        reload(instrument)
        from genx.models.lib import instrument_numba
        self.assertTrue(instrument.SquareIntensity is instrument_numba.SquareIntensity)

    def test_numba_error(self):
        global instrument
        lib.USE_NUMBA=True
        from genx.models.lib import instrument_numba
        # make sure the module misses a required function so an exception is raised in instrument when importing
        del instrument_numba.SquareIntensity
        reload(instrument)
        instrument_numba.SquareIntensity=self.SquareIntensity

if __name__=='__main__':
    unittest.main()



"""
An replacement for wx.SingleChoiceDialog that does not have issues with oversized width
and allows additional styling of elements.
"""

import wx

class SCDialog(wx.Dialog):
    def __init__(self, parent, message, caption, choices,
                 background_colors=None,
                 style=wx.CHOICEDLG_STYLE, pos=wx.DefaultPosition):
        wx.Dialog.__init__(self, parent, title=caption, style=style, pos=pos)
        vbox=wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(vbox)
        vbox.Add(wx.StaticText(self, label=message), 0, wx.FIXED_MINSIZE|wx.ALL, 4)
        self.list_widget=wx.ListBox(self, style=wx.LB_SINGLE|wx.LB_HSCROLL|wx.LB_OWNERDRAW,
                                  choices=choices)
        if background_colors:
            for i, ci in enumerate(background_colors):
                self.list_widget.SetItemBackgroundColour(i, ci)

        vbox.Add(self.list_widget, 1, wx.EXPAND)

        self.list_widget.Bind(wx.EVT_LISTBOX_DCLICK, self.selected)

    def selected(self, event):
        self.EndModal(wx.ID_OK)

    def GetSelection(self):
        return self.list_widget.GetSelection()


import wx
import sys
from logging import debug, error, info, warning


def ShowQuestionDialog(frame, message, title='Question?', yes_no=False):
    exc_info=sys.exc_info()
    debug(message, exc_info=exc_info[0] and exc_info)
    if yes_no:
        flags = wx.YES | wx.NO | wx.YES_DEFAULT | wx.ICON_QUESTION
    else:
        flags = wx.OK | wx.CANCEL | wx.OK_DEFAULT | wx.ICON_QUESTION
    dlg=wx.MessageDialog(frame, message, title, flags)
    result=dlg.ShowModal() in [wx.ID_OK, wx.ID_YES]
    dlg.Destroy()
    return result

def ShowNotificationDialog(frame, message, title='Information'):
    exc_info=sys.exc_info()
    info(message, exc_info=exc_info[0] and exc_info)
    dlg=wx.MessageDialog(frame, message, title,
                         wx.OK | wx.ICON_INFORMATION
                         )
    dlg.ShowModal()
    dlg.Destroy()

def ShowWarningDialog(frame, message, title='Warning'):
    exc_info=sys.exc_info()
    warning(message, exc_info=exc_info[0] and exc_info)
    dlg=wx.MessageDialog(frame, message, title,
                         wx.OK | wx.ICON_ERROR
                         )
    dlg.ShowModal()
    dlg.Destroy()

def ShowErrorDialog(frame, message, title='ERROR'):
    exc_info=sys.exc_info()
    error(message, exc_info=exc_info[0] and exc_info)
    dlg=wx.MessageDialog(frame, message, title,
                         wx.OK | wx.ICON_ERROR
                         )
    dlg.ShowModal()
    dlg.Destroy()




"""
Module to handle different cases of exceptions in the user interface.
"""

import traceback
from logging import debug, warning, error, Handler, Formatter, CRITICAL, LogRecord

import wx

from ..exceptions import GenxError

class CatchModelError:
    """
    A context manager that displays an error dialog if an exception is raised within the 'with' block.
    If the error is a GenX exception the display is slightly different.
    """
    def __init__(self, parent:wx.TopLevelWindow=None,
                 action:str='execution', step:str=None,
                 status_update: callable=None):
        self.parent=parent
        self.action=action
        self.step=step
        self._status_update=status_update
        self.successful=False

    def status_update(self, text):
        if self._status_update:
            self._status_update(text)

    def __enter__(self):
        debug(f'enter {self.action}/{self.step}', stacklevel=3)
        if self.step:
            self.status_update(f'Start {self.step}.')
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is None:
            debug(f'exit {self.action}/{self.step} w/o error', stacklevel=3)
            self.successful=True
            self.status_update(f'Success in {self.step}.')
            return True

        # allow the code to do something with the error, normally not needed
        self.successful=False
        self.exc_type=exc_type
        self.exc_val=exc_val
        self.exc_tb=exc_tb

        message=f"{type(exc_val).__name__} in {self.action}"
        if self.step:
            message+=f", could not {self.step}."
        ext_message=f"{exc_val}"

        if isinstance(exc_val, GenxError):
            warning(f'{message}', exc_info=(exc_type, exc_val, exc_tb), stack_info=True)
            self.status_update(f'Error in {self.step}, {type(exc_val).__name__}')
            title='Warning'
            icon_style=wx.ICON_WARNING
        else:
            error(message, exc_info=(exc_type, exc_val, exc_tb), stack_info=True)
            self.status_update(f'Fatal error in {self.step}, {type(exc_val).__name__}')
            ext_message+='\n\nPython Error (last calls first):\n    '
            ext_message+='\n    '.join(traceback.format_tb(exc_tb)[:2:-1])
            title='Warning'
            icon_style=wx.ICON_ERROR
        full_trace = message+':\n\n'
        full_trace += ''.join(traceback.format_tb(exc_tb))
        full_trace += f'{type(exc_val).__name__}: {exc_val}'

        # make sure the dialog is shown from main thread and after any queued actions
        wx.CallAfter(self.display_message, title, message, ext_message, full_trace, icon_style)
        return True # exception is not raised in main context

    def display_message(self, title, message, ext_message, full_trace, icon_style=wx.ICON_ERROR):
        style=wx.OK|wx.HELP
        dlg = wx.MessageDialog(self.parent, message, title, style|icon_style)
        dlg.SetExtendedMessage(ext_message)
        dlg.SetHelpLabel('Copy to Clipboard')

        result=dlg.ShowModal()
        while result==wx.ID_HELP:
            if wx.TheClipboard.Open():
                wx.TheClipboard.SetData(wx.TextDataObject(full_trace))
                wx.TheClipboard.Close()
            result=dlg.ShowModal()
        dlg.Destroy()

class GuiExceptionHandler(Handler):
    app: wx.App
    show_dialog = True

    def __init__(self, app, level=CRITICAL):
        Handler.__init__(self, level=level)
        self.app=app

    def emit(self, record: LogRecord):
        if not self.show_dialog:
            return
        title = "GenX - Unhandled Python Error"
        message = f"GenX encountered an unexpected error.\n{record.exc_info[0].__name__}: {record.exc_info[1]}"
        ext_message = f"{record.exc_text}"
        ext_message += "\n\nYou can suppress any future warnings by choosing 'Cancel' to close this window."
        full_trace = f"{record.exc_text}"

        self.display_message(title, message, ext_message, full_trace)

    def display_message(self, title, message, ext_message, full_trace, icon_style=wx.ICON_ERROR):
        style=wx.OK|wx.CANCEL|wx.HELP
        dlg = wx.MessageDialog(self.app.GetTopWindow(), message, title, style|icon_style)
        dlg.SetExtendedMessage(ext_message)
        dlg.SetHelpLabel('Copy to Clipboard')

        result=dlg.ShowModal()
        while result==wx.ID_HELP:
            if wx.TheClipboard.Open():
                wx.TheClipboard.SetData(wx.TextDataObject(full_trace))
                wx.TheClipboard.Close()
            result=dlg.ShowModal()
        if result==wx.ID_CANCEL:
            self.show_dialog = False
        dlg.Destroy()



"""
Custom IDs for GenX menus and toolbars.
"""
import wx
from enum import Enum

class ToolId(int, Enum):
    NEW_MODEL=wx.Window.NewControlId()
    NEW_FROM_FILE=wx.Window.NewControlId()
    OPEN_MODEL=wx.Window.NewControlId()
    SAVE_MODEL=wx.Window.NewControlId()
    SIM_MODEL=wx.Window.NewControlId()
    START_FIT=wx.Window.NewControlId()
    STOP_FIT=wx.Window.NewControlId()
    RESTART_FIT=wx.Window.NewControlId()
    CALC_ERROR=wx.Window.NewControlId()
    SOLVER_SELECT=wx.Window.NewControlId()
    ZOOM=wx.Window.NewControlId()
    ERROR_STATS=wx.Window.NewControlId()

class MenuId(int, Enum):
    NEW_MODEL=wx.Window.NewControlId()
    NEW_FROM_FILE=wx.Window.NewControlId()
    OPEN_MODEL=wx.Window.NewControlId()
    SAVE_MODEL=wx.Window.NewControlId()
    SAVE_MODEL_AS=wx.Window.NewControlId()
    MODEL_BATCH=wx.Window.NewControlId()

    IMPORT_DATA=wx.Window.NewControlId()
    IMPORT_TABLE=wx.Window.NewControlId()
    IMPORT_SCRIPT=wx.Window.NewControlId()

    EXPORT_ORSO=wx.Window.NewControlId()
    EXPORT_DATA=wx.Window.NewControlId()
    EXPORT_TABLE=wx.Window.NewControlId()
    EXPORT_SCRIPT=wx.Window.NewControlId()

    PUBLISH_PLOT=wx.Window.NewControlId()
    PRINT_PLOT=wx.Window.NewControlId()
    PRINT_GRID=wx.Window.NewControlId()
    PRINT_SCRIPT=wx.Window.NewControlId()

    QUIT=wx.Window.NewControlId()

    UNDO=wx.Window.NewControlId()
    REDO=wx.Window.NewControlId()
    HISTORY=wx.Window.NewControlId()
    COPY_GRAPH=wx.Window.NewControlId()
    COPY_SIM=wx.Window.NewControlId()
    COPY_TABLE=wx.Window.NewControlId()
    FIND_REPLACE=wx.Window.NewControlId()
    OPEN_IN_EDITOR=wx.Window.NewControlId()

    NEW_DATA=wx.Window.NewControlId()
    DELETE_DATA=wx.Window.NewControlId()
    LOWER_DATA=wx.Window.NewControlId()
    RAISE_DATA=wx.Window.NewControlId()
    TOGGLE_SHOW=wx.Window.NewControlId()
    TOGGLE_USE=wx.Window.NewControlId()
    TOGGLE_ERROR=wx.Window.NewControlId()
    CALCS_DATA=wx.Window.NewControlId()

    TOGGLE_SLIDER=wx.Window.NewControlId()
    ZOOM=wx.Window.NewControlId()
    ZOOM_ALL=wx.Window.NewControlId()
    Y_SCALE_LIN=wx.Window.NewControlId()
    Y_SCALE_LOG=wx.Window.NewControlId()
    X_SCALE_LIN=wx.Window.NewControlId()
    X_SCALE_LOG=wx.Window.NewControlId()
    AUTO_SCALE=wx.Window.NewControlId()
    USE_TOGGLE_SHOW=wx.Window.NewControlId()

    SIM_MODEL=wx.Window.NewControlId()
    EVAL_MODEL=wx.Window.NewControlId()
    TOGGLE_CUDA=wx.Window.NewControlId()
    START_FIT=wx.Window.NewControlId()
    STOP_FIT=wx.Window.NewControlId()
    RESTART_FIT=wx.Window.NewControlId()
    CALC_ERROR=wx.Window.NewControlId()
    ANALYZE=wx.Window.NewControlId()
    AUTO_SIM=wx.Window.NewControlId()

    SET_OPTIMIZER=wx.Window.NewControlId()
    SET_DATA_LOADER=wx.Window.NewControlId()
    SET_IMPORT=wx.Window.NewControlId()
    SET_PLOT=wx.Window.NewControlId()
    SET_PROFILE=wx.Window.NewControlId()
    SET_EDITOR=wx.Window.NewControlId()

    HELP_MODEL=wx.Window.NewControlId()
    HELP_FOM=wx.Window.NewControlId()
    HELP_PLUGINS=wx.Window.NewControlId()
    HELP_DATA_LOADERS=wx.Window.NewControlId()
    HELP_MANUAL=wx.Window.NewControlId()
    HELP_EXAMPLES=wx.Window.NewControlId()
    HELP_HOMEPAGE=wx.Window.NewControlId()
    HELP_ABOUT=wx.Window.NewControlId()
    HELP_DEBUG=wx.Window.NewControlId()

def rebuild_IDs():
    # re-generate above Enums with new IDs for e.g. window recreation
    global ToolId, MenuId
    new_vals = dict([(key, wx.Window.NewControlId()) for key in ToolId.__members__.keys()])
    ToolId = Enum('ToolId', new_vals, type=int, module=__name__)
    new_vals = dict([(key, wx.Window.NewControlId()) for key in MenuId.__members__.keys()])
    MenuId = Enum('MenuId', new_vals, type=int, module=__name__)



'''
Library for combined x-ray and neutrons simulations.
====================================================
The neutron simulations is capable of handling non-magnetic, magnetic
non-spin flip as well as neutron spin-flip reflectivity.

Classes
-------

Layer
~~~~~
``Layer(b = 0.0, d = 0.0, f = 0.0+0.0J, dens = 1.0, magn_ang = 0.0, magn = 0.0, sigma = 0.0)``

``b``
   The neutron scattering length per formula unit in fm (fermi meter =
   1e-15m)
``d``
   The thickness of the layer in AA (Angstroms = 1e-10m)
``f``
   The x-ray scattering length per formula unit in electrons. To be
   strict it is the number of Thompson scattering lengths for each
   formula unit.
``dens``
   The density of formula units in units per Angstroms. Note the units!
``magn_ang``
   The angle of the magnetic moment in degress. 0 degrees correspond to
   a moment collinear with the neutron spin.
``magn``
   The magnetic moment per formula unit (same formula unit as b and dens
   refer to)
``sigma``
   The root mean square roughness of the top interface of the layer in
   Angstroms.
``xs_ai``
   The sum of the absorption cross section and the incoherent scattering
   cross section in barns for neutrons

Stack
~~~~~
``Stack(Layers = [], Repetitions = 1)``

``Layers``
   A ``list`` consiting of ``Layer``s in the stack the first item is
   the layer closest to the bottom
``Repetitions``
   The number of repsetions of the stack

Sample
~~~~~~
``Sample(Stacks = [], Ambient = Layer(), Substrate = Layer())``

``Stacks``
   A ``list`` consiting of ``Stack``s in the stacks the first item is
   the layer closest to the bottom
``Ambient``
   A ``Layer`` describing the Ambient (enviroment above the sample).
   Only the scattering lengths and density of the layer is used.
``Substrate``
   A ``Layer`` describing the substrate (enviroment below the sample).
   Only the scattering lengths, density and roughness of the layer is
   used.

Instrument
~~~~~~~~~~
``Instrument(probe = 'x-ray', wavelength = 1.54, coords = '2θ', I0 = 1.0 res = 0.001, restype = 'no conv', respoints = 5, resintrange = 2, beamw = 0.01, footype = 'no corr', samplelen = 10.0, incangle = 0.0, pol = 'uu')``

``probe``
    Describes the radiation and measurments used is one of: 'x-ray',
    'neutron', 'neutron pol', 'neutron pol spin flip', 'neutron tof',
    'neutron pol tof' or the respective number 0, 1, 2, 3, 4, 5, 6. The
    calculations for x-rays uses ``f`` for the scattering length for
    neutrons ``b`` for 'neutron pol', 'neutron pol spin flip' and 'neutron
    pol tof' alternatives the ``magn`` is used in the calculations. Note
    that the angle of magnetization ``magn_ang`` is only used in the last
    alternative.
``wavelength``
    The wavelength of the radiation given in AA (Angstroms)
``coords``
    The coordinates of the data given to the SimSpecular function. The
    available alternatives are: 'q' or '2θ'. Alternatively the numbers 0 (q)
    or 1 (tth) can be used.
``I0``
    The incident intensity (a scaling factor)
``Ibkg``
    The background intensity. Added as a constant value to the calculated
    reflectivity
``res``
    The resolution of the instrument given in the coordinates of ``coords``.
    This assumes a gaussian resolution function and ``res`` is the standard
    deviation of that gaussian. If ``restype`` has (dx/x) in its name the
    gaussian standard deviation is given by res*x where x is either in tth
    or q.
``restype``
    Describes the rype of the resolution calculated. One of the
    alterantives: 'no conv', 'fast conv', 'full conv and varying res.',
    'fast conv + varying res.', 'full conv and varying res. (dx/x)', 'fast
    conv + varying res. (dx/x)'. The respective numbers 0-3 also works. Note
    that fast convolution only alllows a single value into res wheras the
    other can also take an array with the same length as the x-data (varying
    resolution)
``respoints``
    The number of points to include in the resolution calculation. This is
    only used for 'full conv and vaying res.', 'fast conv + varying res',
    'full conv and varying res. (dx/x)' and 'fast conv + varying res.
    (dx/x)'.
``resintrange``
    Number of standard deviatons to integrate the resolution function times
    the reflectivity over
``footype``
    Which type of footprint correction is to be applied to the simulation.
    One of: 'no corr', 'gauss beam' or 'square beam'. Alternatively, the
    number 0-2 are also valid. The different choices are self expnalatory.
``beamw``
    The width of the beam given in mm. For 'gauss beam' it should be the
    standard deviation. For 'square beam' it is the full width of the beam.
``samplelen``
    The length of the sample given in mm
``incangle``
    The incident angle of the neutrons, only valid in tof mode
``pol``
    The measured polarization of the instrument. Valid options are:
    'uu','dd', 'ud', 'du' or 'ass' the respective number 0-3 also works.
'''
import numpy as np

from .lib import paratt as Paratt
from .lib import neutron_refl as MatrixNeutron
from .lib.instrument import *
from .lib import refl as refl
from .lib.physical_constants import r_e, muB_to_SL


# Preamble to define the parameters needed for the models outlined below:

# import core.paratt as slow_paratt

ModelID = 'SpecNX'
# InstrumentParameters={'Wavelength':1.54, 'Coordinates':1, 'I0':1.0, 'Sim': 0,
#    'Res':0.001, 'Restype':0, 'Respoints':5, 'Resintrange':2, 'Beaw':0.01,
#    'Footype':0.0, 'Samlen':10.0, 'Incangle':0.0}
__pars__ = ['Layer', 'Stack', 'Sample', 'Instrument']

instrument_string_choices = {
    'probe': ['x-ray', 'neutron', 'neutron pol',
              'neutron pol spin flip', 'neutron tof', 'neutron pol tof'],
    'coords': ['q', '2θ', 'tth'],
    'restype': ['no conv', 'fast conv',
                'full conv and varying res.', 'fast conv + varying res.',
                'full conv and varying res. (dx/x)', 'fast conv + varying res. (dx/x)'],
    'footype': ['no corr', 'gauss beam', 'square beam'],
    'pol': ['uu', 'dd', 'ud', 'ass', 'du']
    }
InstrumentParameters = {
    'probe': 'x-ray', 'wavelength': 1.54, 'coords': '2θ',
    'I0': 1.0, 'res': 0.001,
    'restype': 'no conv', 'respoints': 5, 'resintrange': 2.0, 'beamw': 0.01,
    'footype': 'no corr', 'samplelen': 10.0, 'incangle': 0.5, 'pol': 'uu',
    'Ibkg': 0.0, 'tthoff': 0.0
    }
InstrumentGroups = [('General', ['wavelength', 'coords', 'I0', 'Ibkg', 'tthoff']),
                    ('Resolution', ['restype', 'res', 'respoints', 'resintrange']),
                    ('Probe', ['probe', 'pol', 'incangle']),
                    ('Footprint', ['footype', 'beamw', 'samplelen', ]),
                    ]
InstrumentUnits = {
    'probe': '', 'wavelength': 'Å', 'coords': '',
    'I0': 'arb.', 'res': '[coord]',
    'restype': '', 'respoints': 'pts.', 'resintrange': '[coord]', 'beamw': 'mm',
    'footype': '', 'samplelen': 'mm', 'incangle': '°', 'pol': '',
    'Ibkg': 'arb.', 'tthoff': '°',
    '2θ': '°', 'tth': '°', 'q': 'Å$^-1$'
    }
# Coordinates=1 or '2θ' => twothetainput
# Coordinates=0 or 'q'=> Q input
# probe: Type of simulation
#   'x-ray' or 0: X-rays (One output)
#   'neutron'or 1: Neutrons (One output, ignoring magn, magn_ang)
#   'neutron pol' or 2: Neutrons polarized (Two outputs Ruu,Rdd)
#   'neutron pol spin flip' or 3: Neutrons polarized with spin-flip 
#            (Three outputs Ruu,Rdd,Rud=Rdu, ignoring sigma!)
#   'neutron tof' or 4: Neutrons non-polarized TOF, Inc Angle must be set
#   'neutron pol tof'or 5: Neutrons polarized TOF (non-spin flip),
#         Inc Angle must be set
#
# res stddev of resolution
# restype 0 'none': No resolution convlution
#               1 or 'fast': Fast convolution
#               2 or 'full': Full Convolution +varying resolution
#               3 or 'ffull': Fast convolution varying resolution
# respoints Number of points for the convolution only valid for ResolutionType=2
# resintrange Number of standard deviatons to integrate over default 2
# Parameters for footprint coorections
# footype: 0 or 'no corr': No corections for footprint
#            1 or 'gauss beam': Correction for Gaussian beam => Beaw given in mm and stddev
#            2 or 'square beam': Correction for square profile => Beaw given in full width mm
# samlen= Samplelength in mm.

LayerParameters = {
    'sigma': 0.0, 'dens': 1.0, 'd': 0.0, 'f': (1.0+1.0j)*1e-20,
    'b': 0.0+0.0J, 'xs_ai': 0.0, 'magn': 0.0, 'magn_ang': 0.0
    }
LayerUnits = {
    'sigma': 'AA', 'dens': 'at./AA', 'd': 'AA', 'f': 'el./at.',
    'b': 'fm/at.', 'xs_ai': 'barn/at.', 'magn': 'mu_B/at.', 'magn_ang': 'deg.'
    }
LayerGroups = [('Standard', ['f', 'dens', 'd', 'sigma']),
               ('Neutron', ['b', 'xs_ai', 'magn', 'magn_ang'])]
StackParameters = {'Layers': [], 'Repetitions': 1}
SampleParameters = {'Stacks': [], 'Ambient': None, 'Substrate': None}

AA_to_eV = 12398.5
''' Conversion from Angstrom to eV E = AA_to_eV/lamda.'''

q_limit = 1e-10
''' Minimum allowed q-value '''

__xlabel__ = "q [Å$^{-1}$]"
__ylabel__ = "Instnsity [a.u.]"


# A buffer to save previous calculations for spin-flip calculations
class Buffer:
    Ruu = 0
    Rdd = 0
    Rdu = 0
    Rud = 0
    parameters = None
    TwoThetaQz = None


def footprintcorr(Q, instrument):
    foocor = 1.0
    footype = instrument.getFootype()
    beamw = instrument.getBeamw()
    samlen = instrument.getSamplelen()

    if instrument.getProbe() in [instrument_string_choices['probe'][4],
                                 instrument_string_choices['probe'][5], 4, 5]:
        theta = instrument.getIncangle()
        # if ai is an array, make sure it gets repeated for every resolution point
        if type(theta) is ndarray and instrument.getRestype() in [2, 4,
                                                                  instrument_string_choices['restype'][2],
                                                                  instrument_string_choices['restype'][4]]:
            theta = (theta*ones(instrument.getRespoints())[:, newaxis]).flatten()

    else:
        # theta =  180./pi*arcsin(instrument.getWavelength() / 4.0 / pi * Q)
        theta = QtoTheta(instrument.getWavelength(), Q)
    if footype==1 or footype==instrument_string_choices['footype'][1]:
        foocor = GaussIntensity(theta, samlen/2.0, samlen/2.0, beamw)
    elif footype==2 or footype==instrument_string_choices['footype'][2]:
        foocor = SquareIntensity(theta, samlen, beamw)
    elif footype==0 or footype==instrument_string_choices['footype'][0]:
        pass
    else:
        raise ValueError('The choice of footprint correction, footype,'
                         'is WRONG')

    return foocor


def resolutioncorr(R, TwoThetaQz, foocor, instrument, weight):
    ''' Do the convolution of the reflectivity to account for resolution effects.'''
    restype = instrument.getRestype()
    if restype==instrument_string_choices['restype'][1] or restype==1:
        R = ConvoluteFast(TwoThetaQz, R[:]*foocor, instrument.getRes(),
                          range=instrument.getResintrange())
    elif (restype==instrument_string_choices['restype'][2] or restype==2 or
          restype==instrument_string_choices['restype'][4] or restype==4):
        R = ConvoluteResolutionVector(TwoThetaQz, R[:]*foocor, weight)
    elif restype==instrument_string_choices['restype'][3] or restype==3:
        R = ConvoluteFastVar(TwoThetaQz, R[:]*foocor, instrument.getRes(), range=instrument.getResintrange())
    elif restype==instrument_string_choices['restype'][5] or restype==5:
        R = ConvoluteFastVar(TwoThetaQz, R[:]*foocor, instrument.getRes()*TwoThetaQz,
                             range=instrument.getResintrange())
    elif restype==instrument_string_choices['restype'][0] or restype==0:
        R = R[:]*foocor
    else:
        raise ValueError('The choice of resolution type, restype,'
                         'is WRONG')
    return R


def resolution_init(TwoThetaQz, instrument):
    ''' Inits the dependet variable with regards to coordinates and resolution.'''
    global __xlabel__
    __xlabel__ = "q [Å$^{-1}$]"

    restype = instrument.getRestype()
    weight = 0
    if restype==2 or restype==instrument_string_choices['restype'][2]:
        (TwoThetaQz, weight) = ResolutionVector(TwoThetaQz[:],
                                                instrument.getRes(), instrument.getRespoints(),
                                                range=instrument.getResintrange())
    elif restype==4 or restype==instrument_string_choices['restype'][4]:
        (TwoThetaQz, weight) = ResolutionVector(TwoThetaQz[:],
                                                instrument.getRes()*TwoThetaQz, instrument.getRespoints(),
                                                range=instrument.getResintrange())
    # TTH values given as x
    if instrument.getCoords() in [instrument_string_choices['coords'][1],
                                  instrument_string_choices['coords'][2], 1, 2]:
        # Q = 4 * pi / instrument.getWavelength() * sin((TwoThetaQz + instrument.getTthoff()) * pi / 360.0)
        Q = TwoThetatoQ(instrument.getWavelength(), TwoThetaQz+instrument.getTthoff())
        __xlabel__ = "2θ [°]"
    # Q vector given....
    elif instrument.getCoords() in [instrument_string_choices['coords'][0], 0]:
        # if there is no tth offset, nothing to be done for Q
        if instrument.getTthoff()==0:
            Q = TwoThetaQz
        # for tof the q-values are not angles but wavelenght, so tth-offset is a scaling factor
        elif instrument.getProbe() in [instrument_string_choices['probe'][4],
                                       instrument_string_choices['probe'][5], 4, 5]:
            ai = instrument.getIncangle()
            # if ai is an array, make sure it gets repeated for every resolution point
            if type(ai) is ndarray and restype in [2, 4,
                                                   instrument_string_choices['restype'][2],
                                                   instrument_string_choices['restype'][4]]:
                ai = (ai*ones(instrument.getRespoints())[:, newaxis]).flatten()
            Q = TwoThetaQz*(sin((ai+instrument.getTthoff()/2.)*pi/180.)/sin(ai*pi/180.))
        else:
            Q = TwoThetatoQ(instrument.getWavelength(),
                            QtoTheta(instrument.getWavelength(), TwoThetaQz)*2.0+instrument.getTthoff())
    else:
        raise ValueError('The value for coordinates, coords, is WRONG! should be q(0) or tth(1).')
    return Q, TwoThetaQz, weight


def neutron_sld(abs_xs, dens, fb, wl):
    return dens*(wl**2/2/pi*fb-1.0J*abs_xs*wl/4/pi)


def Specular(TwoThetaQz, sample, instrument):
    """ Simulate the specular signal from sample when probed with instrument

    # BEGIN Parameters
    TwoThetaQz data.x
    # END Parameters
    """
    return specular_calcs(TwoThetaQz, sample, instrument, return_int=True)


def SpecularField(TwoThetaQz, sample, instrument):
    """ Simulate the specular signal from sample when probed with instrument

    # BEGIN Parameters
    TwoThetaQz data.x
    # END Parameters
    """
    return specular_calcs(TwoThetaQz, sample, instrument, return_int=False)


def specular_calcs(TwoThetaQz, sample, instrument, return_int=True):
    """ Simulate the specular signal from sample when probed with instrument

    # BEGIN Parameters
    TwoThetaQz data.x
    # END Parameters
    """

    # preamble to get it working with my class interface
    restype = instrument.getRestype()
    Q, TwoThetaQz, weight = resolution_init(TwoThetaQz, instrument)
    # often an issue with resolution etc. so just replace Q values < q_limit
    # if any(Q < q_limit):
    #    raise ValueError('The q vector has to be above %.1e, please verify all your x-axis points fulfill this criterion, including possible resolution smearing.'%q_limit)
    Q = maximum(Q, q_limit)

    ptype = instrument.getProbe()
    pol = instrument.getPol()

    parameters = sample.resolveLayerParameters()
    if ptype==instrument_string_choices['probe'][0] or ptype==0:
        # fb = array(parameters['f'], dtype = complex64)
        e = AA_to_eV/instrument.getWavelength()
        fb = refl.cast_to_array(parameters['f'], e).astype(complex128)
    else:
        fb = array(parameters['b'], dtype=complex128)*1e-5
        abs_xs = array(parameters['xs_ai'], dtype=complex128)*1e-4**2

    dens = array(parameters['dens'], dtype=float64)
    d = array(parameters['d'], dtype=float64)
    magn = array(parameters['magn'], dtype=float64)
    # Transform to radians
    magn_ang = array(parameters['magn_ang'], dtype=float64)*pi/180.0

    sigma = array(parameters['sigma'], dtype=float64)

    if ptype==instrument_string_choices['probe'][0] or ptype==0:
        sld = dens*fb*instrument.getWavelength()**2/2/pi
    else:
        wl = instrument.getWavelength()
        # sld = dens*(wl**2/2/pi*sqrt(fb**2 - (abs_xs/2.0/wl)**2) -
        #                       1.0J*abs_xs*wl/4/pi)
        sld = neutron_sld(abs_xs, dens, fb, wl)
    # Ordinary Paratt X-rays
    if ptype==instrument_string_choices['probe'][0] or ptype==0:
        R = Paratt.ReflQ(Q, instrument.getWavelength(), 1.0-r_e*sld, d, sigma, return_int=return_int)
        # print 2.82e-5*sld
    # Ordinary Paratt Neutrons
    elif ptype==instrument_string_choices['probe'][1] or ptype==1:
        R = Paratt.ReflQ(Q, instrument.getWavelength(), 1.0-sld, d, sigma, return_int=return_int)
    # Ordinary Paratt but with magnetization
    elif ptype==instrument_string_choices['probe'][2] or ptype==2:
        msld = muB_to_SL*magn*dens*instrument.getWavelength()**2/2/pi
        # Polarization uu or ++
        if pol==instrument_string_choices['pol'][0] or pol==0:
            R = Paratt.ReflQ(Q, instrument.getWavelength(), 1.0-sld-msld, d, sigma, return_int=return_int)
        # Polarization dd or --
        elif pol==instrument_string_choices['pol'][1] or pol==1:
            R = Paratt.ReflQ(Q, instrument.getWavelength(), 1.0-sld+msld, d, sigma, return_int=return_int)
        elif pol==instrument_string_choices['pol'][3] or pol==3:
            Rp = Paratt.ReflQ(Q, instrument.getWavelength(), 1.0-sld-msld, d, sigma, return_int=return_int)
            Rm = Paratt.ReflQ(Q, instrument.getWavelength(), 1.0-sld+msld, d, sigma, return_int=return_int)
            R = (Rp-Rm)/(Rp+Rm)

        else:
            raise ValueError('The value of the polarization is WRONG.'
                             ' It should be uu(0) or dd(1)')
    # Spin flip
    elif ptype in [3, instrument_string_choices['probe'][3]]:
        # Check if we have calcluated the same sample previous:
        if Buffer.TwoThetaQz is not None:
            Q_ok = Buffer.TwoThetaQz.shape==Q.shape
            if Q_ok:
                Q_ok = not (Buffer.TwoThetaQz!=Q).any()
        else:
            Q_ok = False
        if Buffer.parameters!=parameters or not Q_ok:
            msld = muB_to_SL*magn*dens*instrument.getWavelength()**2/2/pi
            # renormalize SLDs if ambient layer is not vacuum
            if msld[-1]!=0. or sld[-1]!=0:
                msld -= msld[-1]
                sld -= sld[-1]
            sld_p = sld+msld
            sld_m = sld-msld
            Vp = (2*pi/instrument.getWavelength())**2*(sld_p*(2.+sld_p))  # (1-np**2) - better numerical accuracy
            Vm = (2*pi/instrument.getWavelength())**2*(sld_m*(2.+sld_m))  # (1-nm**2)
            (Ruu, Rdd, Rud, Rdu) = MatrixNeutron.Refl(Q, Vp, Vm, d, magn_ang, sigma, return_int=return_int)
            Buffer.Ruu = Ruu
            Buffer.Rdd = Rdd
            Buffer.Rud = Rud
            Buffer.parameters = parameters.copy()
            Buffer.TwoThetaQz = Q.copy()
        else:
            pass
        # Polarization uu or ++
        if pol==instrument_string_choices['pol'][0] or pol==0:
            R = Buffer.Ruu
        # Polarization dd or --
        elif pol==instrument_string_choices['pol'][1] or pol==1:
            R = Buffer.Rdd
        # Polarization ud or +-
        elif (pol==instrument_string_choices['pol'][2] or pol==2 or
              pol==instrument_string_choices['pol'][4] or pol==4):
            R = Buffer.Rud
        # Calculating the asymmetry ass
        elif pol==instrument_string_choices['pol'][3] or pol==3:
            R = (Buffer.Ruu-Buffer.Rdd)/(Buffer.Ruu+Buffer.Rdd+2*Buffer.Rud)
        else:
            raise ValueError('The value of the polarization is WRONG.'
                             ' It should be uu(0), dd(1) or ud(2)')

    # tof
    elif ptype==instrument_string_choices['probe'][4] or ptype==4:
        ai = instrument.getIncangle()
        # if ai is an array, make sure it gets repeated for every resolution point
        if type(ai) is ndarray and restype in [2, 4,
                                               instrument_string_choices['restype'][2],
                                               instrument_string_choices['restype'][4]]:
            ai = (ai*ones(instrument.getRespoints())[:, newaxis]).flatten()
        else:
            ai = ai*ones(Q.shape)
        wl = 4*pi*sin(ai*pi/180)/Q
        sld = neutron_sld(abs_xs[:, newaxis], dens[:, newaxis], fb[:, newaxis], wl)
        R = Paratt.Refl_nvary2(ai, wl,
                               1.0-sld, d, sigma, return_int=return_int)
    # tof spin polarized
    elif ptype==instrument_string_choices['probe'][5] or ptype==5:
        wl = 4*pi*sin(instrument.getIncangle()*pi/180)/Q
        sld = neutron_sld(abs_xs[:, newaxis], dens[:, newaxis], fb[:, newaxis], wl)
        msld = muB_to_SL*magn[:, newaxis]*dens[:, newaxis]*(4*pi*sin(instrument.getIncangle()*pi/180)/Q)**2/2/pi
        # polarization uu or ++
        if pol==instrument_string_choices['pol'][0] or pol==0:
            R = Paratt.Refl_nvary2(instrument.getIncangle()*ones(Q.shape),
                                   (4*pi*sin(instrument.getIncangle()*pi/180)/Q), 1.0-sld-msld, d, sigma,
                                   return_int=return_int)
        # polarization dd or --
        elif pol==instrument_string_choices['pol'][1] or pol==1:
            R = Paratt.Refl_nvary2(instrument.getIncangle()*ones(Q.shape),
                                   (4*pi*sin(instrument.getIncangle()*pi/180)/Q), 1.0-sld+msld, d, sigma,
                                   return_int=return_int)
        # Calculating the asymmetry
        elif pol==instrument_string_choices['pol'][3] or pol==3:
            Rd = Paratt.Refl_nvary2(instrument.getIncangle()*ones(Q.shape),
                                    (4*pi*sin(instrument.getIncangle()*pi/180)/Q),
                                    1.0-sld+msld, d, sigma, return_int=return_int)
            Ru = Paratt.Refl_nvary2(instrument.getIncangle()*ones(Q.shape),
                                    (4*pi*sin(instrument.getIncangle()*pi/180)/Q),
                                    1.0-sld-msld, d, sigma, return_int=return_int)
            R = (Ru-Rd)/(Ru+Rd)

        else:
            raise ValueError('The value of the polarization is WRONG.'
                             ' It should be uu(0) or dd(1) or ass')
    else:
        raise ValueError('The choice of probe is WRONG')
    if return_int:
        # FootprintCorrections
        foocor = footprintcorr(Q, instrument)
        # Resolution corrections
        R = resolutioncorr(R, TwoThetaQz, foocor, instrument, weight)

        return R*instrument.getI0()+instrument.getIbkg()
    else:
        return R


def EnergySpecular(Energy, TwoThetaQz, sample, instrument):
    ''' Simulate the specular signal from sample when probed with instrument. Energy should be in eV.

    # BEGIN Parameters
    Energy data.x
    TwoThetaQz 3.0
    # END Parameters
    '''
    # preamble to get it working with my class interface
    restype = instrument.getRestype()
    # TODO: Fix so that resolution can be included.
    if restype!=0 and restype!=instrument_string_choices['restype'][0]:
        raise ValueError('Only no resolution is allowed for energy scans.')

    wl = AA_to_eV/Energy
    global __xlabel__
    __xlabel__ = 'E [eV]'

    # TTH values given as x
    if instrument.getCoords()==instrument_string_choices['coords'][1] \
            or instrument.getCoords()==1:
        theta = TwoThetaQz/2.0
    # Q vector given....
    elif instrument.getCoords()==instrument_string_choices['coords'][0] \
            or instrument.getCoords()==0:
        theta = arcsin(TwoThetaQz*wl/4/pi)*180.0/pi

    else:
        raise ValueError('The value for coordinates, coords, is WRONG!'
                         'should be q(0) or tth(1).')
    Q = 4*pi/wl*sin((2*theta+instrument.getTthoff())*pi/360.0)

    ptype = instrument.getProbe()

    parameters = sample.resolveLayerParameters()
    if ptype==instrument_string_choices['probe'][0] or ptype==0:
        fb = refl.cast_to_array(parameters['f'], Energy).astype(complex128)
    else:
        fb = array(parameters['b'], dtype=complex128)*1e-5
        abs_xs = array(parameters['xs_ai'], dtype=complex128)*1e-4**2

    dens = array(parameters['dens'], dtype=float64)
    d = array(parameters['d'], dtype=float64)
    sigma = array(parameters['sigma'], dtype=float64)

    if ptype==instrument_string_choices['probe'][0] or ptype==0:
        sld = dens[:, newaxis]*fb*wl**2/2/pi
    else:
        wl = instrument.getWavelength()
        sld = dens*(wl**2/2/pi*sqrt(fb**2-(abs_xs/2.0/wl)**2)-1.0J*abs_xs*wl/4/pi)
    # Ordinary Paratt X-rays
    if ptype==instrument_string_choices['probe'][0] or ptype==0:
        # R = Paratt.ReflQ(Q,instrument.getWavelength(),1.0-2.82e-5*sld,d,sigma)
        R = Paratt.Refl_nvary2(theta, wl, 1.0-r_e*sld, d, sigma)
    else:
        raise ValueError('The choice of probe is WRONG')
    # TODO: Fix corrections
    # FootprintCorrections
    # foocor = footprintcorr(Q, instrument)
    # Resolution corrections
    # R = resolutioncorr(R, TwoThetaQz, foocor, instrument, weight)

    return R*instrument.getI0()+instrument.getIbkg()


def OffSpecular(TwoThetaQz, ThetaQx, sample, instrument):
    ''' Function that simulates the off-specular signal (not implemented)
    
    # BEGIN Parameters
    TwoThetaQz 1.0
    ThetaQx data.x
    # END Parameters
    '''
    raise NotImplementedError('Not implemented use model interdiff insteads')
    return TwoThetaQz, ThetaQx


def SLD_calculations(z, item, sample, inst):
    ''' Calculates the scatteringlength density as at the positions z
    if item is None or "all" the function returns a dictonary of values.
    Otherwise it returns the item as identified by its string.
    
    # BEGIN Parameters
    z data.x
    item 'Re'
    # END Parameters
    '''
    parameters = sample.resolveLayerParameters()
    if hasattr(sample, 'crop_sld') and sample.crop_sld!=0:
        crop_top_bottom = abs(sample.crop_sld)
        inter = dict([(key, 0.) for key in parameters])
        if sample.crop_sld>0:
            inter['d'] = sum(parameters['d'][crop_top_bottom:-crop_top_bottom])
        else:
            inter['d'] = 5.0
            inter['dens'] = 0.1
            inter['b'] = 12.0+0j
            inter['f'] = 100.0+0j
        if len(parameters['dens'])>2*crop_top_bottom:
            for key, value in parameters.items():
                val_start = value[:crop_top_bottom]
                val_end = value[-crop_top_bottom:]
                parameters[key] = val_start+[inter[key]]+val_end
    dens = array(parameters['dens'], dtype=float32)
    # f = array(parameters['f'], dtype = complex64)
    e = AA_to_eV/inst.getWavelength()
    f = refl.cast_to_array(parameters['f'], e).astype(complex64)
    b = array(parameters['b'], dtype=complex64)*1e-5
    abs_xs = array(parameters['xs_ai'], dtype=float32)*1e-4**2
    wl = inst.getWavelength()
    ptype = inst.getProbe()
    magnetic = False
    mag_sld = 0
    sld_unit = r'r_{e}/\AA^{3}'
    if ptype==instrument_string_choices['probe'][0] or ptype==0:
        sld = dens*f
    elif ptype==instrument_string_choices['probe'][1] or ptype==1 or \
            ptype==instrument_string_choices['probe'][4] or ptype==4:
        sld = dens*(wl**2/2/pi*b-1.0J*abs_xs*wl/4/pi)/1e-6/(wl**2/2/pi)
        sld_unit = r'10^{-6}\AA^{-2}'
    else:
        magnetic = True
        sld = dens*(wl**2/2/pi*b-1.0J*abs_xs*wl/4/pi)/1e-6/(wl**2/2/pi)
        magn = array(parameters['magn'], dtype=float64)
        # Transform to radians
        magn_ang = array(parameters['magn_ang'], dtype=float64)*pi/180.0
        mag_sld = 2.645*magn*dens*10.
        mag_sld_x = mag_sld*cos(magn_ang)
        mag_sld_y = mag_sld*sin(magn_ang)
        sld_unit = r'10^{-6}\AA^{-2}'

    d = array(parameters['d'], dtype=float64)
    d = d[1:-1]
    # Include one extra element - the zero pos (substrate/film interface)
    int_pos = cumsum(r_[0, d])
    sigma = array(parameters['sigma'], dtype=float64)[:-1]+1e-7
    if z is None:
        z = arange(-sigma[0]*5, int_pos.max()+sigma[-1]*5, 0.5)
    if not magnetic:
        rho = sum((sld[:-1]-sld[1:])*(0.5-
                                      0.5*erf((z[:, newaxis]-int_pos)/sqrt(2.)/sigma)), 1)+sld[-1]
        dic = {
            'Re': real(rho), 'Im': imag(rho), 'z': z,
            'SLD unit': sld_unit
            }
    else:
        sld_p = sld+mag_sld
        sld_m = sld-mag_sld
        rho_p = sum((sld_p[:-1]-sld_p[1:])*(0.5-
                                            0.5*erf((z[:, newaxis]-int_pos)/sqrt(2.)/sigma)), 1)+sld_p[-1]
        rho_m = sum((sld_m[:-1]-sld_m[1:])*(0.5-
                                            0.5*erf((z[:, newaxis]-int_pos)/sqrt(2.)/sigma)), 1)+sld_m[-1]
        rho_nucl = (rho_p+rho_m)/2.
        if (magn_ang!=0.).any():
            rho_mag_x = sum((mag_sld_x[:-1]-mag_sld_x[1:])*
                            (0.5-0.5*erf((z[:, newaxis]-int_pos)/sqrt(2.)/sigma)), 1)+mag_sld_x[-1]
            rho_mag_y = sum((mag_sld_y[:-1]-mag_sld_y[1:])*
                            (0.5-0.5*erf((z[:, newaxis]-int_pos)/sqrt(2.)/sigma)), 1)+mag_sld_y[-1]
            dic = {
                'Re non-mag': real(rho_nucl), 'Im non-mag': imag(rho_nucl),
                'mag': real(rho_p-rho_m)/2, 'z': z, 'mag_x': rho_mag_x, 'mag_y': rho_mag_y,
                'SLD unit': sld_unit
                }
        else:
            dic = {
                'Re non-mag': real(rho_nucl), 'Im non-mag': imag(rho_nucl),
                'mag': real(rho_p-rho_m)/2, 'z': z,
                'SLD unit': sld_unit
                }
    if item is None or item=='all':
        return dic
    else:
        try:
            return dic[item]
        except:
            raise ValueError('The chosen item, %s, does not exist'%item)

POL_CHANNELS = ['uu', 'ud', 'du', 'dd']

def PolSpecular(TwoThetaQz, p1, p2, F1, F2, sample, instrument):
    """
    Specular reflectivity of polarized measurement with finite polarization.
    The polarization parameters are in accordance to the definition used in
    A.R. Wildes publication, Review of Scientific Instruments 70, 11 (1999)
    https://doi.org/10.1063/1.1150060

    pol_params = (p1, p2, f1, f2)
    p1: polarizer efficiency, 0: 100% spin-up, 1: 100% spin-down, 0.5: unpolarized
    p2: analyzer efficiency,  0: 100% spin-up, 1: 100% spin-down, 0.5: unpolairzed
    F1/F2: Flipper efficienty, 0: 100% efficient, 1: no flipping

    # BEGIN Parameters
    TwoThetaQz data.x
    p1 0.
    p2 0.
    F1 0.
    F2 0.
    # END Parameters

    """
    inst_pol = instrument.pol
    if not inst_pol in POL_CHANNELS:
        raise ValueError(f"Instrument polarization as to be one of {POL_CHANNELS}.")
    if instrument.probe not in [3, instrument_string_choices['probe'][3]]:
        raise ValueError("Polarization corrected simulation requires probe to be 'neutron pol spin flip'")

    instrument.pol = 'uu'
    uu = Specular(TwoThetaQz, sample, instrument)
    instrument.pol = 'dd'
    dd = Specular(TwoThetaQz, sample, instrument)
    instrument.pol = 'ud'
    ud = Specular(TwoThetaQz, sample, instrument)
    du = ud
    instrument.pol = inst_pol

    P = get_pol_matrix(p1, p2, F1, F2)
    Pline = P[POL_CHANNELS.index(instrument.pol)]
    I = Pline[:, newaxis] * np.vstack([uu, ud, du, dd])
    return I.sum(axis=0)

SimulationFunctions = {
    'Specular': Specular,
    'PolSpecular': PolSpecular,
    'SpecularField': SpecularField,
    'OffSpecular': OffSpecular,
    'SLD': SLD_calculations,
    'EnergySpecular': EnergySpecular,
    }

(Instrument, Layer, Stack, Sample) = refl.MakeClasses(InstrumentParameters,
                                                      LayerParameters, StackParameters, SampleParameters,
                                                      SimulationFunctions,
                                                      ModelID)

if __name__=='__main__':
    from .utils import UserVars, fp, fw


    # BEGIN Instrument DO NOT CHANGE
    inst = Instrument(footype='gauss beam', probe='x-ray', beamw=0.04, resintrange=2, pol='uu', wavelength=1.54,
                      respoints=5, Ibkg=0.0, I0=2, samplelen=10.0, restype='no conv', coords='2θ', res=0.001,
                      incangle=0.0)
    fp.set_wavelength(inst.wavelength)
    # Compability issues for pre-fw created gx files
    try:
        fw
    except:
        pass
    else:
        fw.set_wavelength(inst.wavelength)
    # END Instrument

    # BEGIN Sample DO NOT CHANGE
    Amb = Layer(b=0, d=0.0, f=(1e-20+1e-20j), dens=1.0, magn_ang=0.0, sigma=0.0, xs_ai=0.0, magn=0.0)
    topPt = Layer(b=0, d=11.0, f=fp.Pt, dens=4/3.92**3, magn_ang=0.0, sigma=3.0, xs_ai=0.0, magn=0.0)
    TopFe = Layer(b=0, d=11.0, f=fp.Fe, dens=2/2.866**3, magn_ang=0.0, sigma=2.0, xs_ai=0.0, magn=0.0)
    Pt = Layer(b=0, d=11.0, f=fp.Pt, dens=4/3.92**3, magn_ang=0.0, sigma=2.0, xs_ai=0.0, magn=0.0)
    Fe = Layer(b=0, d=11, f=fp.Fe, dens=2/2.866**3, magn_ang=0.0, sigma=2.0, xs_ai=0.0, magn=0.0)
    bufPt = Layer(b=0, d=45, f=fp.Pt, dens=4/3.92**3, magn_ang=0.0, sigma=2, xs_ai=0.0, magn=0.0)
    bufFe = Layer(b=0, d=2, f=fp.Fe, dens=2/2.866**3, magn_ang=0.0, sigma=2, xs_ai=0.0, magn=0.0)
    Sub = Layer(b=0, d=0.0, f=fp.Mg+fp.O, dens=2/4.2**3, magn_ang=0.0, sigma=4.0, xs_ai=0.0, magn=0.0)

    ML = Stack(Layers=[Fe, Pt], Repetitions=100)

    sample = Sample(Stacks=[ML], Ambient=Amb, Substrate=Sub)
    # END Sample

    # BEGIN Parameters DO NOT CHANGE
    cp = UserVars()
    cp.new_var('Lambda', 33)
    # END Parameters

    import cProfile


    tth = np.arange(0.01, 8.0, 0.01)
    cProfile.run('[sample.SimSpecular(tth, inst) for i in range(10)]', sort='tottime')


    def Sim(data):
        I = []
        # BEGIN Dataset 0 DO NOT CHANGE
        TopFe.setD(Fe.d)
        TopFe.setSigma(Fe.sigma)
        Pt.setD(cp.Lambda-Fe.d)
        I.append(sample.SimSpecular(data[0].x, inst))
        # END Dataset 0
        return I



'''
All custom events used in GenX to be directly imported.
'''
import wx
from wx.lib.newevent import NewEvent
from ..core.decorators import decorator

# a decorator to make sure events processed by a method are skiped first
@decorator
def skips_event(function, *args, **kwargs):
    event=args[1]
    event.Skip()
    return function(*args, **kwargs)

## main_window
# Generating an event type:
new_model_type=wx.NewEventType()
# Creating an event binder object
EVT_NEW_MODEL=wx.PyEventBinder(new_model_type)
# Generating an event type:
sim_plot_type=wx.NewEventType()
# Creating an event binder object
EVT_SIM_PLOT=wx.PyEventBinder(sim_plot_type)


## parametergrid
# Event for when the grid has new values
grid_change, EVT_PARAMETER_GRID_CHANGE=NewEvent()
# Event for then the value of a parameter has changed. Should be used to do
# simulations interactively.
value_change, EVT_PARAMETER_VALUE_CHANGE=NewEvent()
set_parameter_value, EVT_PARAMETER_SET_VALUE=NewEvent()
move_parameter, EVT_MOVE_PARAMETER=NewEvent()
inset_parameter, EVT_INSERT_PARAMETER=NewEvent()
delete_parameters, EVT_DELETE_PARAMETERS=NewEvent()
sort_and_group_parameters, EVT_SORT_AND_GROUP_PARAMETERS=NewEvent()

## datalist
# Generating an event type:
data_list_type=wx.NewEventType()
# Creating an event binder object
EVT_DATA_LIST=wx.PyEventBinder(data_list_type)
# Send when plot setting is to be changed
update_plotsettings, EVT_UPDATE_PLOTSETTINGS=NewEvent()

## plotpanel
# Event for a click inside an plot which yields a number
plot_position, EVT_PLOT_POSITION=NewEvent()
# Event to tell the main window that the zoom state has changed
state_changed, EVT_PLOT_SETTINGS_CHANGE=NewEvent()

## solvergui
# Custom events needed for updating and message parsing between the different
# modules.
update_script, EVT_UPDATE_SCRIPT=NewEvent()
update_plot, EVT_UPDATE_PLOT=NewEvent()
update_text, EVT_SOLVER_UPDATE_TEXT=NewEvent()
update_parameters, EVT_UPDATE_PARAMETERS=NewEvent()
fitting_ended, EVT_FITTING_ENDED=NewEvent()
autosave, EVT_AUTOSAVE=NewEvent()
batch_next, EVT_BATCH_NEXT=NewEvent()

## plugins
# new model is ready with a script as value.
update_model_event, EVT_UPDATE_MODEL=wx.lib.newevent.NewEvent()

## custom_logging
log_message_event, EVT_LOG_MESSAGE=wx.lib.newevent.NewEvent()


"""
An interface to show output messages from python logging with convenient fielters.
"""

import wx
from traceback import format_exception
from typing import List, Tuple
import logging

from.custom_events import log_message_event, EVT_LOG_MESSAGE


class GuiHandler(logging.Handler, wx.EvtHandler):
    def __init__(self, level=logging.DEBUG):
        logging.Handler.__init__(self, level=level)
        wx.EvtHandler.__init__(self)
        formatter = logging.Formatter('[%(levelname)s] - %(asctime)s - %(filename)s:%(lineno)i:%(funcName)s %(message)s',
                                      datefmt='%H:%M:%S')
        self.setFormatter(formatter)

    def emit(self, record: logging.LogRecord):
        fmt_message = self.format(record)
        evt=log_message_event(record=record, fmt_message=fmt_message)
        wx.QueueEvent(self, evt)

class RecordDisplayDialog(wx.Dialog):
    def __init__(self, parent, record: logging.LogRecord):
        wx.Dialog.__init__(self, parent, style=wx.DEFAULT_DIALOG_STYLE|wx.STAY_ON_TOP|wx.RESIZE_BORDER,
                           name=f'GenX log message')
        self.SetTitle(f'GenX {record.levelname} message')
        vbox=wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(vbox)
        vbox.Add(wx.StaticText(self, label=f'Severity: {record.levelname}'))
        vbox.Add(wx.StaticText(self, label=f'Process: {record.processName}'))
        vbox.Add(wx.StaticText(self, label=f'Thread: {record.threadName}'))
        vbox.Add(wx.StaticText(self, label=f'Module: {record.module}'))
        vbox.Add(wx.StaticText(self, label=f'Line: {record.lineno}'))
        vbox.Add(wx.StaticText(self, label=f'\nMessage:'))
        msg=wx.TextCtrl(self, value=record.message, style=wx.TE_READONLY|wx.TE_MULTILINE|wx.TE_DONTWRAP)
        font=msg.GetFont()
        font.SetFamily(wx.FONTFAMILY_TELETYPE)
        msg.SetFont(font)
        vbox.Add(msg, 0, wx.EXPAND)
        if record.exc_info:
            vbox.Add(wx.StaticText(self, label='\nError:'))
            exc_message=''.join(format_exception(*record.exc_info))
            if record.stack_info:
                exc_message+='\n'+record.stack_info
            msg=wx.TextCtrl(self, value=exc_message, style=wx.TE_READONLY|wx.TE_MULTILINE|wx.TE_DONTWRAP)
            font=msg.GetFont()
            font.SetFamily(wx.FONTFAMILY_TELETYPE)
            msg.SetFont(font)
            vbox.Add(msg, 1, wx.EXPAND)

class LoggingDialog(wx.Dialog):
    logged_events: List[logging.LogRecord]

    def __init__(self, parent=None):
        wx.Dialog.__init__(self, parent, style=wx.DEFAULT_DIALOG_STYLE|wx.STAY_ON_TOP|wx.RESIZE_BORDER,
                           name='GenX Log')
        self.SetTitle('GenX Logging History')
        self.handler = GuiHandler()
        logging.getLogger().addHandler(self.handler)
        self.build_layout()
        self.logged_events=[]
        self.log_level=logging.INFO

        self.Bind(wx.EVT_CLOSE, self.close_window)
        self.Bind(wx.EVT_LIST_ITEM_ACTIVATED, self.show_event_message, self.log_list)
        self.handler.Bind(EVT_LOG_MESSAGE, self.OnMessage)

        pos=parent.GetPosition()
        size=parent.GetSize()
        self.SetSize(400, size.height)
        self.SetPosition(wx.Point(pos.x+size.width, pos.y))

    def build_layout(self):
        vbox=wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(vbox)
        topbox=wx.BoxSizer(wx.HORIZONTAL)
        vbox.Add(topbox)
        rb=wx.RadioButton(self, -1, 'DEBUG')
        self.Bind(wx.EVT_RADIOBUTTON, lambda evt: self.update_loglevel(logging.DEBUG), rb)
        topbox.Add(rb)
        rb=wx.RadioButton(self, -1, 'INFO')
        rb.SetValue(True)
        self.Bind(wx.EVT_RADIOBUTTON, lambda evt: self.update_loglevel(logging.INFO), rb)
        topbox.Add(rb)
        rb=wx.RadioButton(self, -1, 'WARNING')
        self.Bind(wx.EVT_RADIOBUTTON, lambda evt: self.update_loglevel(logging.WARNING), rb)
        topbox.Add(rb)
        rb=wx.RadioButton(self, -1, 'ERROR')
        self.Bind(wx.EVT_RADIOBUTTON, lambda evt: self.update_loglevel(logging.ERROR), rb)
        topbox.Add(rb)

        self.log_list=wx.ListCtrl(self, style=wx.LC_REPORT|wx.LC_SINGLE_SEL)
        self.log_list.AppendColumn('Time')
        self.log_list.AppendColumn('Level')
        self.log_list.AppendColumn('Message')
        # self.log_list.setResizeColumn(3)
        vbox.Add(self.log_list, 1, wx.EXPAND)

    def close_window(self, event):
        # make sure the handler is removed on close of dialog
        event.Skip()
        logging.getLogger().removeHandler(self.handler)

    @property
    def filtered_events(self):
        return [ri for ri in self.logged_events if ri.levelno>=self.log_level]

    def update_loglevel(self, level):
        self.log_level=level
        self.log_list.DeleteAllItems()
        for i, record in enumerate(self.filtered_events):
            if (i+1)<len(self.logged_events) and \
                    self.logged_events[i+1].message==record.message and \
                    self.logged_events[i+1].lineno==record.lineno:
                continue
            self.append_event(record)
        self.resize_columns()

    def show_event_message(self, event: wx.ListEvent):
        record=self.logged_events[event.GetData()]
        RecordDisplayDialog(self, record).Show()

    def append_event(self, record: logging.LogRecord, new=False):
        if record.levelno>=self.log_level:
            if new:
                prev_event_evts=[ri for ri in self.logged_events if ri.levelno>=self.log_level]
                if len(prev_event_evts)>1 and \
                    prev_event_evts[-2].message==record.message and \
                    prev_event_evts[-2].lineno==record.lineno:
                    # don't repeat the same error to speed up diaplsy
                    return
            self.log_list.Append((record.asctime, record.levelname, record.message.splitlines()[0]))
            last = self.log_list.GetItemCount()-1
            self.log_list.SetItemData(last, self.logged_events.index(record))

    def resize_columns(self):
        self.log_list.SetColumnWidth(0, wx.LIST_AUTOSIZE)
        self.log_list.SetColumnWidth(1, wx.LIST_AUTOSIZE)
        self.log_list.SetColumnWidth(2, wx.LIST_AUTOSIZE)
        items=self.log_list.GetItemCount()
        if items>0:
            self.log_list.EnsureVisible(items-1)

    def OnMessage(self, evt):
        self.logged_events.append(evt.record)
        self.append_event(evt.record, new=True)
        self.resize_columns()


"""
A dialog that acts as file-like object for console output.
"""

import wx

from io import StringIO

class TextOutputDialog(wx.Dialog):
    def __init__(self, parent, autoflush=True):
        wx.Dialog.__init__(self, parent, style=wx.DEFAULT_DIALOG_STYLE|wx.RESIZE_BORDER,
                           name=f'GenX action details')
        self.SetTitle(f'Text output')
        vbox=wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(vbox)
        self.msg=wx.TextCtrl(self, value="", style=wx.TE_READONLY|wx.TE_MULTILINE|wx.TE_DONTWRAP)
        font=self.msg.GetFont()
        font.SetFamily(wx.FONTFAMILY_TELETYPE)
        self.msg.SetFont(font)
        vbox.Add(self.msg, 1, wx.EXPAND)
        self.text=StringIO()
        self.autoflush=autoflush

    def write(self, txt):
        self.text.write(txt)
        if self.autoflush:
            self.flush()

    def writable(self):
        return True

    def writelines(self, lines):
        self.text.writelines(lines)
        if self.autoflush:
            self.flush()

    def flush(self):
        self.msg.SetValue(self.text.getvalue())
        self.msg.SetScrollPos( wx.VERTICAL, self.msg.GetScrollRange(wx.VERTICAL))
        self.msg.SetInsertionPoint(-1)
        wx.YieldIfNeeded()



"""
Helper module to support ORSO file header information to build reflectivity models.
"""

from dataclasses import dataclass
from copy import deepcopy
from typing import List, Optional, TYPE_CHECKING

from orsopy import fileio

from ...add_ons.help_modules.materials_db import Formula, MASS_DENSITY_CONVERSION

if TYPE_CHECKING:
    # make sure the wx based plugins don't need to be imported at runtime
    from ...add_ons.SimpleReflectivity import Plugin as SRPlugin
    from ...add_ons.Reflectivity import Plugin as RPlugin

@dataclass
class InstrumentInformation:
    probe: str
    coords: str
    wavelength: float

@dataclass
class LayerData:
    dens: float
    formula: Formula
    f: complex
    b: complex
    d: float
    sigma: float

@dataclass
class LayerModel:
    substrate: LayerData
    ambient: LayerData
    stacks: List[List[LayerData]]
    repetitions: List[int]

class OrsoHeaderAnalyzer:
    header: fileio.Orso
    instrument: InstrumentInformation
    layer_model: Optional[LayerModel] = None

    def __init__(self, meta):
        self.model = None
        meta = deepcopy(meta)
        self.header = fileio.Orso(**meta)
        self.analyze_instrument_info()
        if hasattr(fileio, 'model_language') and self.header.data_source.sample.model:
            self.analyze_model()

    @property
    def instrument_settings(self) -> fileio.InstrumentSettings:
        return self.header.data_source.measurement.instrument_settings

    @property
    def experiment(self) -> fileio.Experiment:
        return self.header.data_source.experiment

    def get_probe(self):
        if self.experiment.probe=='neutron':
            pol = self.instrument_settings.polarization or 'unpolarized'
            if pol=='unpolarized':
                probe = 'neutron'
            elif pol in ['pp', 'mm', 'pm', 'mp']:
                probe = 'neutron pol spin flip'
            else:
                probe = 'neutron pol'
        else:
            probe = 'x-ray'
        return probe

    def analyze_instrument_info(self):
        probe = self.get_probe()

        # detect x-axis unit
        if self.header.columns[0].unit in [None, '1/angstrom', '1/nm']:
            coords = 'q'
        else:
            # not officially ORSO confirm
            coords = '2θ'
        try:
            wavelength = float(self.instrument_settings.wavelength.as_unit('angstrom'))
        except Exception:
            wavelength = 1.54
        if not isinstance(wavelength, float):
            # in case wavelength is a ValueRange (ToF)
            wavelength = 1.0
        self.instrument = InstrumentInformation(probe=probe, coords=coords, wavelength=wavelength)

    @staticmethod
    def get_layer_data(layer) -> LayerData:
        layer.material.generate_density()
        if layer.material.formula:
            formula = Formula.from_str(layer.material.formula)
            if layer.material.number_density:
                dens = layer.material.number_density.as_unit('1/angstrom**3')
            elif layer.material.mass_density:
                dens = layer.material.mass_density.as_unit('g/cm**3')* \
                                 MASS_DENSITY_CONVERSION/formula.mFU()
            else:
                from genx.models.utils import bc
                try:
                    dens = layer.material.get_sld().real/eval(formula.b()).real
                except Exception:
                    dens = 0.1
            b = formula.b()
            f = formula.f()
        else:
            formula = None
            b = layer.material.get_sld()
            f = layer.material.get_sld()
            dens = 0.1
        d = layer.thickness.as_unit('angstrom')
        sigma = layer.roughness.as_unit('angstrom')
        return LayerData(dens=dens, formula=formula, b=b, f=f, d=d, sigma=sigma)


    def analyze_model(self):
        from orsopy.fileio import model_language
        header_model = self.header.data_source.sample.model

        res_stack = header_model.resolve_stack()
        stacks = []
        repetitions = []
        last_stack = True
        for si in res_stack:
            if hasattr(si, 'repetitions'):
                repetitions.append(si.repetitions)
                stack_layers = []
                for stack_item in si.sequence:
                    if isinstance(stack_item, model_language.Layer):
                        stack_layers.append(self.get_layer_data(stack_item))
                    else:
                        stack_layers += list(map(self.get_layer_data, stack_item.resolve_to_layers()))
                stacks.append(stack_layers)
                last_stack = True
            elif last_stack:
                last_stack = False
                repetitions.append(1)
                stacks.append([self.get_layer_data(si)])
            else:
                stacks[-1].append(self.get_layer_data(si))

        # make sure we remove ambient and substrate from a single repetition stack
        if repetitions[0]!=1:
            repetitions.insert(0, 1)
            repetitions[1] -= 1
            stacks.insert(0, stacks[0])
        if repetitions[-1]!=1:
            repetitions[-1] -= 1
            repetitions.append(1)
            stacks.append(stacks[-1])
        ambient = stacks[0].pop(0)
        substrate = stacks[-1].pop(-1)
        for i, si in reversed(list(enumerate(stacks))):
            if len(si)==0:
                stacks.pop(i)
                repetitions.pop(i)
        self.layer_model = LayerModel(ambient=ambient, substrate=substrate, stacks=stacks, repetitions=repetitions)

    def simple_refl_layer(self, layer: LayerData, pos=1):
        if layer.formula:
            dens = layer.dens*layer.formula.mFU()/MASS_DENSITY_CONVERSION
            return [None, "Formula", layer.formula,
                    False, str(dens), False, '0.0',
                    False, str(layer.d), False, str(layer.sigma),
                    pos]
        else:
            return [None, "Formula", "SLD",
                    False, str(layer.b.real if self.instrument.probe!='x-ray' else layer.f.real), False, '0.0',
                    False, str(layer.d), False, str(layer.sigma),
                    pos]

    def build_simple_model(self, refl: "SRPlugin"):
        from ...add_ons.SimpleReflectivity import TOP_LAYER, ML_LAYER, BOT_LAYER
        refl.sample_widget.sample_table.ResetModel()
        refl.sample_widget.inst_params['probe'] = self.instrument.probe
        refl.sample_widget.inst_params['wavelength'] = self.instrument.wavelength
        refl.sample_widget.inst_params['coords'] = self.instrument.coords
        refl.sample_widget.inst_params['res'] = 0.01
        if self.layer_model:
            # if any stack is repeated, use that as central one
            if any([ri>1 for ri in self.layer_model.repetitions]):
                pos = TOP_LAYER
                is_ML = True
            else:
                pos = ML_LAYER
                is_ML = False

            ID = 0
            layers = []
            repetitions = 1
            for i, (ri, si) in enumerate(zip(self.layer_model.repetitions, self.layer_model.stacks)):
                if is_ML and pos==ML_LAYER:
                    pos = BOT_LAYER
                if ri>1 and pos==TOP_LAYER:
                    pos = ML_LAYER
                    repetitions = ri
                    for li in si:
                        nl = self.simple_refl_layer(li, pos=pos)
                        nl[0] = f'Layer_{ID:02}'
                        layers.append(nl)
                        ID+=1
                else:
                    for li in ri*si:
                        nl = self.simple_refl_layer(li, pos=pos)
                        nl[0] = f'Layer_{ID:02}'
                        layers.append(nl)
                        ID+=1
            refl.sample_widget.sample_table.ambient = self.simple_refl_layer(self.layer_model.ambient)
            refl.sample_widget.sample_table.substrate = self.simple_refl_layer(self.layer_model.substrate)
            refl.sample_widget.sample_table.RebuildTable(layers)
            refl.sample_widget.sample_table.repetitions = repetitions
        refl.sample_widget.last_sample_script = refl.sample_widget.sample_table.getModelCode()
        refl.sample_widget.UpdateModel(re_color=True)

    def build_reflectivity(self, refl: "RPlugin"):
        refl.CreateNewModel('models.spec_nx')
        # detect source radiation
        inst = refl.sample_widget.instruments['inst']
        inst.probe = self.instrument.probe
        inst.coords = self.instrument.coords
        inst.wavelength = self.instrument.wavelength
        pol_names = {
            'po': 'uu', 'mo': 'dd', 'op': 'uu', 'om': 'dd',
            'pp': 'uu', 'mm': 'dd', 'pm': 'ud', 'mp': 'du'
            }
        # set resolution column
        if len(self.header.columns)>3 and self.header.columns[3].error_of==self.header.columns[0].name:
            inst.restype = 'full conv and varying res.'
            inst.respoints = 7
            inst.resintrange = 2.5
            el = refl.simulation_widget.GetExpressionList()
            for i, data_item in enumerate(refl.GetModel().data):
                if len(data_item.meta['columns'])>3 and (
                        data_item.meta['columns'][3].get('error_of', None)==
                        data_item.meta['columns'][0]['name']):
                    el[i].append(f'inst.setRes(data[{i}].res)')
                else:
                    el[i].append(f'inst.setRes(0.001)')
                # set polarization channel
                pol = data_item.meta['data_source']['measurement']['instrument_settings'].get('polarization',
                                                                                              'unpolarized')
                if pol in pol_names:
                    el[i].append(f'inst.setPol("{pol_names[pol]}")')

        if self.layer_model:
            lm = self.layer_model
            tmp = refl.sampleh.sample.Ambient
            tmp.b, tmp.f, tmp.dens = lm.ambient.b, lm.ambient.f, lm.ambient.dens
            tmp = refl.sampleh.sample.Substrate
            tmp.b, tmp.f, tmp.dens, tmp.sigma = lm.substrate.b, lm.substrate.f, lm.substrate.dens, lm.substrate.sigma

            pos = 1
            for si, (rep, stack) in enumerate(zip(lm.repetitions, lm.stacks)):
                refl.sampleh.insertItem(pos, 'Stack', f'ST_{si}')
                stack_obj = refl.sampleh.sample.Stacks[0]
                stack_obj.Repetitions = rep
                pos += 1
                for li, layer in enumerate(stack):
                    refl.sampleh.insertItem(pos, 'Layer', f'L_{si}_{li:02}')
                    pos += 1

                    layer_obj = stack_obj.Layers[0]

                    layer_obj.b, layer_obj.f, layer_obj.dens = layer.b, layer.f, layer.dens
                    layer_obj.d , layer_obj.sigma = layer.d, layer.sigma

            refl.sample_widget.Update()

        refl.WriteModel()

# include GenX materials in density resolution
try:
    from orsopy.fileio import model_language
except ImportError:
    pass
else:
    from orsopy.utils.density_resolver import DensityResolver
    from orsopy.utils.resolver_slddb import ResolverSLDDB
    from orsopy.utils.chemical_formula import Formula as OrsoFormula
    from ...add_ons.help_modules.materials_db import mdb

    class ResolverGenX(DensityResolver):
        comment = "from GenX "

        def resolve_formula(self, formula: OrsoFormula) -> float:
            frm = Formula.from_str(str(formula))
            if frm not in mdb:
                raise ValueError(f"Could not find material {formula}")
            else:
                res_formula, density = mdb[frm]
                self.comment = f"density from GenX db {res_formula}"
                return 1e3*eval(density)

        def resolve_elemental(self, formula: OrsoFormula) -> float:
            raise ValueError("GenX resolve only used for known materials.")

    model_language.DENSITY_RESOLVERS = [ResolverGenX(), ResolverSLDDB()]



"""
Generate a simple test class that is used to test the hdf interface of all important
data types.
"""

import unittest
import tempfile
import h5py
import numpy as np
from dataclasses import dataclass, field

from genx.core.h5_support import H5HintedExport

def default_dict():
    return {'str': 'a', 'int': 13, 'float': 134.2, 'complex': 41.5+3j, 'subdict': {'a': 'a'}}

@dataclass
class H5TestSub(H5HintedExport):
    h5group_name = 'subtest'

    testvalue: str

@dataclass
class H5Tester(H5HintedExport):
    h5group_name = 'test_subgroup'
    # some attributes (without type hint) will not be saved
    non_saved = None
    non_saved_int = 13
    non_saved_float = 134.5

    saved_str: str
    saved_int: int
    saved_float: float
    saved_complex: complex
    saved_dict: dict
    saved_array: np.ndarray
    saved_subclass: H5TestSub

    saved_default_str: str = 'default'
    saved_default_int: int = 13
    saved_default_float: float = 31.45
    saved_default_complex: complex = 13.5+3j
    saved_default_dict: dict = field(default_factory=default_dict)
    saved_default_array: np.ndarray = field(default_factory=lambda: np.arange(100))


class TestH5HintedExport(unittest.TestCase):

    def test_write_read(self):
        ds1 = H5Tester('abc1', 10, 45.3, 85.4+5.3j, default_dict(), np.linspace(-1.,1.,50), H5TestSub('test'))
        ds2 = H5Tester('abc2', 20, 25.3, 45.4+2.3j, {'b': 'c', 'd': 13}, np.linspace(-2,1,50), H5TestSub('test'),
                       'def', 42, 42.5, 0.4-2j, {'b': 'c', 'd': 13}, np.linspace(-2,1,50),) # overwritten defaults
        ds1re = H5Tester('', 0, 0.0, 0j, {}, np.array([]), H5TestSub(''))
        ds2re = H5Tester('', 0, 0.0, 0j, {}, np.array([]), H5TestSub(''), '', 0, 0.0, 0j, {}, np.array([]))

        with tempfile.TemporaryFile('w+b') as tmp:
            hdf = h5py.File(tmp, 'w')
            g1 = hdf.create_group('ds1')
            ds1.write_h5group(g1)
            g2 = hdf.create_group('ds2')
            ds2.write_h5group(g2)
            hdf.close()

            hdf = h5py.File(tmp, 'r')
            ds1re.read_h5group(hdf['ds1'])
            ds2re.read_h5group(hdf['ds2'])

        # remove array objects for comparison
        ds1a = ds1.saved_array
        ds1da = ds1.saved_default_array
        ds1.saved_array = None
        ds1.saved_default_array = None

        ds2a = ds2.saved_array
        ds2da = ds2.saved_default_array
        ds2.saved_array = None
        ds2.saved_default_array = None

        ds1ra = ds1re.saved_array
        ds1rda = ds1re.saved_default_array
        ds1re.saved_array = None
        ds1re.saved_default_array = None

        ds2ra = ds2re.saved_array
        ds2rda = ds2re.saved_default_array
        ds2re.saved_array = None
        ds2re.saved_default_array = None

        self.assertEqual(ds1, ds1re)
        self.assertEqual(ds2, ds2re)

        np.testing.assert_array_equal(ds1a, ds1ra)
        np.testing.assert_array_equal(ds1da, ds1rda)
        np.testing.assert_array_equal(ds2a, ds2ra)
        np.testing.assert_array_equal(ds2da, ds2rda)



'''
===================================
:mod:`seifert_nja` NJA ASCII format
===================================

Loads the data from *.nja ASCII files with limited header analysis.
'''

import numpy as np
from ..data_loader_framework import Template
from ..utils import ShowWarningDialog


try:
    import wx
except ImportError:
    # in case of console usinge withou wx beeing installed put a mock class/module
    class void():
        pass


    wx = void()
    wx.Dialog = void


class Plugin(Template):
    wildcard='*.nja'

    def __init__(self, parent):
        Template.__init__(self, parent)

    def CanOpen(self, file_path):
        if not Template.CanOpen(self, file_path):
            return False
        with open(file_path, 'r', encoding='utf-8') as fh:
            l1=fh.readline()
            l2=fh.readline()
        return l2.startswith('#ScanTableParameter')

    def evaluate_header(self, header):
        output = {}
        for line in header:
            if not line.startswith('&'):
                continue
            items = line.split('&')
            if items[1].startswith('ScanAxis'):
                output['ScanAxis'] = items[1].split('=')[1].strip()
            elif items[1].startswith('Axis'):
                output[items[1]]=dict([tuple(ei.strip().split('=', 2)) for ei in items[1:]])
            else:
                for ei in items[1:]:
                    output[ei.split('=')[0]] = ei.split('=',2)[-1].strip()

        return output

    def LoadData(self, dataset, filename, data_id=0):
        '''LoadData(self, dataset, filename) --> none

        Loads the data from filename into the dataset object.
        '''
        try:
            with open(filename, encoding='utf-8', errors='ignore') as fh:
                header = [fh.readline()]
                while not header[-1].startswith('&NoValues='):
                    header.append(fh.readline())
                load_array = np.loadtxt(fh, delimiter=None)
        except Exception as e:
            ShowWarningDialog(self.parent, 'Could not load the file: '+ \
                              filename+' \nPlease check the format.\n\n numpy.loadtxt' \
                              +' gave the following error:\n'+str(e))
        else:
            # For the freak case of only one data point
            if len(load_array.shape)<2:
                load_array = np.array([load_array])
            # Check so we have enough columns
            if load_array.shape[1]<2:
                ShowWarningDialog(self.parent, 'The data file does not contain enough number of columns. It has '
                                  +str(load_array.shape[1])+' columns. Rember that the column index start at zero!')
                self.SetStatusText('Could not load data - not enough columns')
                return

            # The data is set by the default Template.__init__ function
            # Know the loaded data goes into *_raw so that they are not
            # changed by the transforms
            dataset.x_raw = load_array[:, 0]
            dataset.y_raw = load_array[:, 1]

            header = self.evaluate_header(header)
            if header.get('ScanAxis', '') == "O":
                dataset.x_command = '2*x'
            else:
                dataset.x_command = 'x'


            # Check if we have errors in the data - if not handle it with nan's
            dataset.error_raw = np.sqrt(load_array[:, 1])

            # Run the commands on the data - this also sets the x,y, error members of the data item.
            dataset.run_command()

            # insert metadata into ORSO compatible fields
            dataset.meta['data_source']['facility'] = 'Seifert nja loader'
            dataset.meta['data_source']['experiment']['probe']='xray'
            dataset.meta['data_source']['measurement']['scheme']='angle-dispersive'
            # header evaluations
            inst = dataset.meta['data_source']['measurement']['instrument_settings']
            inst['incident_angle'] = {'min': float(dataset.x_raw.min()), 'max': float(dataset.x_raw.max()),
                                      'unit': 'degree'}
            inst['wavelength'] = {'magnitude': 1.54, 'unit': 'angstrom'}
            # insert all metadata into user field
            dataset.meta['nja_header'] = header



"""
Tests of low lever functionality of GenX Model class.
"""
import os
import unittest
import h5py
import tempfile
from pickle import loads, dumps

from genx.model import Model


class TestModelClass(unittest.TestCase):
    m: Model

    def setUp(self):
        self.m = Model()
        self.example_path = os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))),
                                         'genx', 'examples')

    def test_save_load_gx(self):
        fd, fname = tempfile.mkstemp(suffix='.gx')
        self.m.save(fname)
        self.m.load(fname)
        os.close(fd)

        self.assertEqual(self.m.filename, fname)
        self.assertTrue(self.m.saved)

        self.m = Model()

    def test_load_h5(self):
        fname = os.path.join(self.example_path, 'SuperAdam_SiO.hgx')
        f = h5py.File(fname.encode('utf-8'), 'r')
        g = f[self.m.h5group_name]
        self.m.read_h5group(g)

        self.assertTrue(self.m.saved)
        self.assertLess(0, len(self.m.data[0].x_raw))
        self.assertLess(3, len(self.m.parameters))
        self.assertNotEqual(self.m.script.strip(), '')

        self.m = Model()

    def test_save_h5(self):
        fd, fname = tempfile.mkstemp(suffix='.gx')
        f = h5py.File(fname.encode('utf-8'), 'w')
        g = f.create_group(self.m.h5group_name)
        self.m.write_h5group(g)
        f.close()
        os.close(fd)
        self.m = Model()

    def test_pickle(self):
        pstr = dumps(self.m)
        remodel = loads(pstr)
        for attr in ['opt', 'script', 'data', 'parameters']:
            with self.subTest(f'attr={attr}'):
                old = getattr(self.m, attr)
                new = getattr(remodel, attr)
                self.assertEqual(old, new)


if __name__=='__main__':
    unittest.main()



import warnings
from dataclasses import dataclass
from logging import debug

import matplotlib
import matplotlib.axes
from matplotlib.backends.backend_wxagg import FigureCanvasWxAgg
from matplotlib.figure import Figure

from numpy import sin, cos, tan, pi, exp, log, log10, sqrt, array, arange, hstack, vstack, linspace, arcsin, arccos, arctan
import wx
from wx.py.editwindow import EditWindow
from ..core.config import BaseConfig, Configurable
from ..data import DataList, DataSet


# ==============================================================================
class SimplePlotPanel(wx.Panel):
    ax: matplotlib.axes.Axes

    def __init__(self, parent, id=-1, color=None, dpi=None,
                 style=wx.NO_FULL_REPAINT_ON_RESIZE | wx.EXPAND | wx.ALL, **kwargs):
        wx.Panel.__init__(self, parent, id=id, style=style, **kwargs)
        if dpi is None:
            dpi = wx.GetApp().dpi_scale_factor*96.  # wx.GetDisplayPPI()[0]
        self.parent = parent
        debug('init PlotPanel - setup figure')
        self.figure = Figure(figsize=(1.0, 1.0), dpi=dpi)
        debug('init PlotPanel - setup canvas')
        self.canvas = FigureCanvasWxAgg(self, -1, self.figure)
        self.SetColor(color)
        self._resizeflag = True
        self.print_size = (15./2.54, 12./2.54)
        # self._SetSize()

        # debug('init PlotPanel - bind events')
        self.Bind(wx.EVT_IDLE, self._onIdle)
        self.Bind(wx.EVT_SIZE, self._onSize)

        cursor = wx.Cursor(wx.CURSOR_CROSS)
        self.canvas.SetCursor(cursor)
        self.ax = self.figure.add_subplot(111)

        # Create the drawing bitmap
        self.bitmap = wx.Bitmap(1, 1, depth=wx.BITMAP_SCREEN_DEPTH)
        #        DEBUG_MSG("__init__() - bitmap w:%d h:%d" % (w,h), 2, self)
        self._isDrawn = False
        debug('end init PlotPanel')

    def clear(self):
        self.figure.clear()
        self.ax = self.figure.add_subplot(111)

    def SetColor(self, rgbtuple=None):
        ''' Set the figure and canvas color to be the same '''
        if not rgbtuple:
            rgbtuple = self.parent.GetBackgroundColour()
            # wx.SystemSettings.GetColour(wx.SYS_COLOUR_BTNFACE).Get()
        col = [c/255. for c in rgbtuple]
        self.figure.set_facecolor(col)
        self.figure.set_edgecolor(col)
        self.canvas.SetBackgroundColour(wx.Colour(*rgbtuple))

    def _onSize(self, evt):
        self._resizeflag = True
        self._SetSize()

    def _onIdle(self, evt):
        if self._resizeflag:
            self._resizeflag = False
            self._SetSize()

    def _SetSize(self):
        ''' This method can be called to force the Plot to be a desired
            size which defaults to the ClientSize of the Panel.
        '''
        pixels = self.GetMinSize()

        self.canvas.SetSize(pixels)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", UserWarning)
            try:
                self.figure.tight_layout(h_pad=0)
            except ValueError:
                pass

    def flush_plot(self):
        # with warnings.catch_warnings():
        #     warnings.simplefilter("ignore", UserWarning)
        #     self.figure.tight_layout(h_pad=0)
        self.canvas.draw()


@dataclass
class PublicationConfig(BaseConfig):
    section = 'publication graph'

    width: float = 3.39
    heigth: float = 2.36
    font_size: int = int(matplotlib.rcParams['font.size'])
    font_familty: int = wx.FONTFAMILY_DEFAULT
    font_face: str = matplotlib.rcParams['font.sans-serif'][0]

    start_text: str = repr('''fig.set_facecolor('white')
# for other matplotlib rc style options see https://matplotlib.org/stable/tutorials/introductory/customizing.html
rcParams['font.family'] = 'sans-serif'
rcParams['font.size'] = font_size # dialog entry
rcParams['font.sans-serif'] = font_face # dialog entry
rcParams['font.weight'] = font_weight # dialog entry
rcParams['mathtext.fontset'] = 'dejavusans' # 'dejavuserif'|'cm'|'stix'|'stixsans'
clear()

for di in data:
    if not di.show:
        continue
    errorbar(di.x, di.y, yerr=di.error, label="data-"+di.name, **di.data_kwds)
    semilogy(di.x, di.y_sim, label="sim-"+di.name, **di.sim_kwds)
xlabel(model.__xlabel__)
ylabel(model.__ylabel__)
legend()
tight_layout(pad=0.5)
show()
#savefig(r"your_file_name.png", dpi=300)'''.splitlines())


class PublicationDialog(wx.Dialog, Configurable):
    opt: PublicationConfig

    def __init__(self, parent, id=-1, data: DataList = None, module=None):
        wx.Dialog.__init__(self, parent, id=id,
                           style=wx.MAXIMIZE_BOX | wx.MINIMIZE_BOX | wx.RESIZE_BORDER | wx.DEFAULT_DIALOG_STYLE)
        Configurable.__init__(self)
        self.ReadConfig()
        self.SetTitle('Custom plotting for Publication')

        self.data = data
        self.module = module
        self.dpi_scale_factor = parent.dpi_scale_factor
        hbox = wx.BoxSizer(wx.HORIZONTAL)
        self.SetSizer(hbox)
        vbox = wx.BoxSizer(wx.VERTICAL)
        hbox.Add(vbox, 10, wx.EXPAND | wx.ALL, 4)
        self.tinput = EditWindow(self, wx.ID_ANY)
        self.tinput.SetBackSpaceUnIndents(True)
        self.tinput.SetText('\n'.join(eval(self.opt.start_text)))

        vbox.Add(self.tinput, 1, wx.EXPAND, 0)
        button = wx.Button(self, label='plot')
        vbox.Add(button, 0, wx.EXPAND, 0)

        vbox2 = wx.BoxSizer(wx.VERTICAL)
        hbox.Add(vbox2, 1, wx.EXPAND | wx.ALL, 4)
        self.plot = SimplePlotPanel(self)
        self.plot.SetMinSize(wx.Size(300, 300))
        vbox2.Add(self.plot, 1, wx.EXPAND, 0)
        self.error_text = wx.StaticText(self)
        self.error_text.SetBackgroundColour(wx.Colour(255, 150, 150))
        self.error_text.Hide()
        vbox2.Add(self.error_text, 0, wx.EXPAND, 0)
        hbox2 = wx.BoxSizer(wx.HORIZONTAL)
        vbox2.Add(hbox2, 0, wx.EXPAND, 0)
        self.width = wx.SpinCtrlDouble(self, min=0.1, max=30.0, initial=self.opt.width,
                                       inc=0.01, style=wx.SP_ARROW_KEYS)
        self.height = wx.SpinCtrlDouble(self, min=0.1, max=30.0, initial=self.opt.heigth,
                                        inc=0.01, style=wx.SP_ARROW_KEYS)
        hbox2.Add(wx.StaticText(self, label='Width (inches)'))
        hbox2.Add(self.width, 0, wx.FIXED_MINSIZE | wx.RIGHT, 10)
        hbox2.Add(wx.StaticText(self, label='Height (inches)'))
        hbox2.Add(self.height, 0, wx.FIXED_MINSIZE | wx.RIGHT, 10)
        hbox2 = wx.BoxSizer(wx.HORIZONTAL)
        vbox2.Add(hbox2, 0, wx.EXPAND, 0)
        self.font = wx.FontPickerCtrl(self, style=wx.FNTP_FONTDESC_AS_LABEL)
        self.font.SetSelectedFont(wx.Font(self.opt.font_size, self.opt.font_familty,
                                          wx.FONTSTYLE_NORMAL, wx.FONTWEIGHT_NORMAL, False,
                                          faceName=self.opt.font_face))
        hbox2.Add(wx.StaticText(self, label='Font'))
        hbox2.Add(self.font, 0, wx.EXPAND, 0)

        hbox2 = wx.BoxSizer(wx.HORIZONTAL)
        vbox2.Add(hbox2, 0, wx.EXPAND, 0)
        save_button = wx.Button(self, label='store current script and settings as default')
        hbox2.Add(save_button, 0, wx.EXPAND, 0)

        self.hbox = hbox
        self.Bind(wx.EVT_BUTTON, self.OnPlot, button)
        self.Bind(wx.EVT_BUTTON, self.OnStoreDefaults, save_button)
        self.Bind(wx.EVT_COMMAND_ENTER, self.OnPlot)
        self.Bind(wx.EVT_SPINCTRLDOUBLE, self.OnPlot)
        self.SetSize(wx.Size(800, 800))

    def OnPlot(self, event):
        txt = self.tinput.GetText()
        font = self.font.GetSelectedFont()
        weights = {wx.FONTWEIGHT_BOLD: 'bold', wx.FONTWEIGHT_NORMAL: 'normal'}
        env = dict(
            rcParams=matplotlib.rcParams,
            data=self.data,
            model=self.module,
            show=self.plot.flush_plot,
            fig=self.plot.figure,
            canvas=self.plot.canvas,
            savefig=self.plot.figure.savefig,
            tight_layout=self.plot.figure.tight_layout,

            font_size=font.GetPointSize(),
            font_face=font.GetFaceName(),
            font_weight=weights[font.GetWeight()],
            )

        def clear():
            self.plot.clear()
            ax = self.plot.ax
            env['plot'] = ax.plot,
            env['semilogy'] = ax.semilogy
            env['errorbar'] = ax.errorbar
            env['title'] = ax.set_title

            env['xlabel'] = ax.set_xlabel
            env['ylabel'] = ax.set_ylabel
            env['legend'] = ax.legend
            env['axes'] = ax

        env['clear'] = clear

        w = int(self.width.GetValue()*72*self.dpi_scale_factor*2)
        h = int(self.height.GetValue()*72*self.dpi_scale_factor*2)
        self.plot.figure.dpi = int(72*self.dpi_scale_factor*2)
        size = wx.Size(w, h)
        self.plot.SetMinSize(size)
        self.plot.SetSize(size)
        clear()
        try:
            exec(txt, env)
        except Exception as e:
            self.error_text.SetLabel(e.__class__.__name__+': '+str(e))
            self.error_text.Show()
        else:
            self.error_text.Hide()

        # make sure we reset the default rc values
        matplotlib.rcdefaults()
        self.hbox.SetItemMinSize(self.plot, size)
        self.Layout()

    def OnStoreDefaults(self, event):
        self.opt.start_text = repr(self.tinput.GetText().splitlines())
        self.opt.width = self.width.GetValue()
        self.opt.heigth = self.height.GetValue()

        font = self.font.GetSelectedFont()
        self.opt.font_size = font.GetPointSize()
        self.opt.font_face = font.GetFaceName()
        self.opt.font_familty = font.GetFamily()
        self.WriteConfig(default=True)



"""
Simple dialog window to show a plot graph.
"""

import wx

from .plotpanel import PlotPanel, BasePlotConfig


class DialogPlotConfig(BasePlotConfig):
    section = 'dialog plot'


class DialogPlot(PlotPanel):

    def __init__(self, parent, id=-1, color=None, dpi=None
                 , style=wx.NO_FULL_REPAINT_ON_RESIZE, **kwargs):
        PlotPanel.__init__(self, parent, id, color, dpi, DialogPlotConfig, style, **kwargs)
        self.update(None)
        self.ax = self.figure.add_subplot(111)
        self.ax.set_autoscale_on(False)


class PlotDialog(wx.Dialog):

    def __init__(self, parent, title='GenX Plot'):
        wx.Dialog.__init__(self, parent, style=wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER | wx.MAXIMIZE_BOX)
        vbox = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(vbox)
        self.SetTitle(title)

        self._plot_dialog = DialogPlot(self)
        vbox.Add(self._plot_dialog, proportion=1, flag=wx.EXPAND)
        self.SetMinSize(wx.Size(200,200))

    @property
    def plot(self):
        return self._plot_dialog.ax.plot

    @property
    def semilogy(self):
        return self._plot_dialog.ax.semilogy

    def draw(self):
        self._plot_dialog.AutoScale()
        return self._plot_dialog.flush_plot()

    @property
    def legend(self):
        return self._plot_dialog.ax.legend

    def clear(self):
        return self._plot_dialog.ax.clear()

    def clear_data(self):
        ax = self._plot_dialog.ax
        while len(ax.lines)>0:
            ax.lines[0].remove()
        while len(ax.collections)>0:
            ax.collections[0].remove()

    @property
    def set_xlabel(self):
        return self._plot_dialog.ax.set_xlabel

    @property
    def set_ylabel(self):
        return self._plot_dialog.ax.set_ylabel

    @property
    def set_xlim(self):
        return self._plot_dialog.ax.set_xlim

    @property
    def set_ylim(self):
        return self._plot_dialog.ax.set_ylim
